"""Module for retrieving node information from API."""
from typing import TYPE_CHECKING, Optional

from .api_event import ApiEvent
from .frames import (
    FrameBase, FrameStatusRequestConfirmation, FrameStatusRequestNotification,
    FrameStatusRequestRequest)
from .session_id import get_new_session_id

if TYPE_CHECKING:
    from pyvlx import PyVLX


class StatusRequest(ApiEvent):
    """Class for retrieving node informationfrom API."""

    def __init__(self, pyvlx: "PyVLX", node_id: int):
        """Initialize SceneList class."""
        super().__init__(pyvlx=pyvlx)
        self.node_id = node_id
        self.success = False
        self.notification_frame: Optional[FrameStatusRequestNotification] = None
        self.session_id: Optional[int] = None

    async def handle_frame(self, frame: FrameBase) -> bool:
        """Handle incoming API frame, return True if this was the expected frame."""
        if (
                isinstance(frame, FrameStatusRequestConfirmation)
                and frame.session_id == self.session_id
        ):
            # We are still waiting for StatusRequestNotification
            return False
        if (
                isinstance(frame, FrameStatusRequestNotification)
                and frame.session_id == self.session_id
        ):
            self.notification_frame = frame
            self.success = True
            return True
        return False

    def request_frame(self) -> FrameStatusRequestRequest:
        """Construct initiating frame."""
        self.session_id = get_new_session_id()
        return FrameStatusRequestRequest(
            session_id=self.session_id,
            node_ids=[self.node_id]
        )



"""Unit tests for FrameSetUTCRequest."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import (
    FrameStatusRequestConfirmation, FrameStatusRequestRequest)
from pyvlx.api.frames.frame_status_request import (
    FrameStatusRequestNotification, StatusRequestStatus)


class TestFrameStatusRequestRequest(unittest.TestCase):
    """Test class FrameStatusRequestRequest."""

    # pylint: disable=too-many-public-methods,invalid-name

    EXAMPLE_FRAME = b"\x00\x1d\x03\x05\x00\xab\x02\x01\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00" \
                    b"\x00\x00\x00\x00\x01\xfe\x00N"

    def test_bytes(self) -> None:
        """Test FrameStatusRequestRequest with nodes 1,2 and session_id 0xAB."""
        frame = FrameStatusRequestRequest(node_ids=[1, 2], session_id=0xAB)
        self.assertEqual(bytes(frame), self.EXAMPLE_FRAME)

    def test_frame_from_raw(self) -> None:
        """Test parse FrameStatusRequestRequest from raw."""
        frame = frame_from_raw(self.EXAMPLE_FRAME)
        self.assertTrue(isinstance(frame, FrameStatusRequestRequest))

    def test_str(self) -> None:
        """Test string representation of FrameStatusRequestRequest."""
        frame = FrameStatusRequestRequest(node_ids=[1, 2], session_id=0xAB)
        self.assertEqual(str(frame), "<FrameStatusRequestRequest session_id=\"171\" node_ids=\"[1, 2]\" "
                                     "status_type=\"StatusType.REQUEST_CURRENT_POSITION\" fpi1=\"254\" fpi2=\"0\"/>")


class TestFrameStatusRequestConfirmation(unittest.TestCase):
    """Test class FrameStatusRequestConfirmation."""

    # pylint: disable=too-many-public-methods,invalid-name

    EXAMPLE_FRAME = b"\x00\x06\x03\x06\x00\xab\x01\xa9"

    def test_bytes(self) -> None:
        """Test FrameStatusRequestConfirmation with session_id 0xAB and status ACCEPTED."""
        frame = FrameStatusRequestConfirmation(session_id=0xAB, status=StatusRequestStatus.ACCEPTED)
        self.assertEqual(bytes(frame), self.EXAMPLE_FRAME)

    def test_frame_from_raw(self) -> None:
        """Test parse FrameStatusRequestConfirmation from raw."""
        frame = frame_from_raw(self.EXAMPLE_FRAME)
        self.assertTrue(isinstance(frame, FrameStatusRequestConfirmation))

    def test_str(self) -> None:
        """Test string representation of FrameStatusRequestConfirmation."""
        frame = FrameStatusRequestConfirmation(session_id=0xAB, status=StatusRequestStatus.ACCEPTED)
        self.assertEqual(str(frame),
                         "<FrameStatusRequestConfirmation session_id=\"171\" status=\"StatusRequestStatus.ACCEPTED\"/>")


class TestFrameStatusRequestNotification(unittest.TestCase):
    """Test class FrameStatusRequestNotification."""

    # pylint: disable=too-many-public-methods,invalid-name

    EXAMPLE_FRAME_EMPTY = b"\x00>\x03\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00" \
                          b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00" \
                          b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00:"
    EXAMPLE_FRAME = b"\x00>\x03\x07\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00" \
                    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00" \
                    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00: "

    def test_bytes(self) -> None:
        """Test FrameStatusRequestNotification."""
        frame = FrameStatusRequestNotification()
        self.assertEqual(bytes(frame), self.EXAMPLE_FRAME_EMPTY)

    def test_frame_from_raw(self) -> None:
        """Test parse FrameStatusRequestNotification from raw."""
        frame = frame_from_raw(self.EXAMPLE_FRAME_EMPTY)
        self.assertTrue(isinstance(frame, FrameStatusRequestNotification))

    def test_str(self) -> None:
        """Test string representation of FrameStatusRequestNotification."""
        frame = FrameStatusRequestNotification()
        self.assertEqual(str(frame), "<FrameStatusRequestNotification session_id=\"0\" "
                                     "status_id=\"0\" node_id=\"0\" run_status=\"RunStatus.EXECUTION_COMPLETED\" "
                                     "status_reply=\"StatusReply.UNKNOWN_STATUS_REPLY\" "
                                     "status_type=\"StatusType.REQUEST_TARGET_POSITION\" status_count=\"0\" "
                                     "parameter_data=\"\"/>")



"""Unit tests for FrameSetNodeNameRequest."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameSetNodeNameRequest


class TestFrameSetNodeNameRequest(unittest.TestCase):
    """Test class FrameSetNodeNameRequest."""

    # pylint: disable=too-many-public-methods,invalid-name

    EXAMPLE_FRAME = (
        b"\x00D\x02\x08\x04Fnord\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x1b"
    )

    def test_bytes(self):
        """Test FrameSetNodeNameRequest with NO_TYPE."""
        frame = FrameSetNodeNameRequest(node_id=4, name="Fnord")
        self.assertEqual(bytes(frame), self.EXAMPLE_FRAME)

    def test_frame_from_raw(self):
        """Test parse FrameSetNodeNameRequest from raw."""
        frame = frame_from_raw(self.EXAMPLE_FRAME)
        self.assertTrue(isinstance(frame, FrameSetNodeNameRequest))
        self.assertEqual(frame.node_id, 4)
        self.assertEqual(frame.name, "Fnord")

    def test_str(self):
        """Test string representation of FrameSetNodeNameRequest."""
        frame = FrameSetNodeNameRequest(node_id=4, name="Fnord")
        self.assertEqual(
            str(frame), '<FrameSetNodeNameRequest node_id="4" name="Fnord"/>'
        )



"""Module for lights."""
from typing import TYPE_CHECKING, Optional

from .api import CommandSend
from .node import Node
from .parameter import Intensity

if TYPE_CHECKING:
    from pyvlx import PyVLX


class DimmableDevice(Node):
    """Meta class for turning on device with one main parameter for intensity."""

    def __init__(self, pyvlx: "PyVLX", node_id: int, name: str, serial_number: Optional[str]):
        """Initialize turning on device.

        Parameters:
            * pyvlx: PyVLX object
            * node_id: internal id for addressing nodes.
                Provided by KLF 200 device
            * name: node name
            * serial_number: serial number of the node.

        """
        super().__init__(
            pyvlx=pyvlx, node_id=node_id, name=name, serial_number=serial_number
        )
        self.intensity = Intensity()

    async def set_intensity(self, intensity: Intensity, wait_for_completion: bool = True) -> None:
        """Set light to desired intensity.

        Parameters:
            * intensity: Intensity object containing the target intensity.
                0% means off, 100% means fully on.
            * wait_for_completion: If set, function will return
                after device has reached target intensity.

        """
        command = CommandSend(
            pyvlx=self.pyvlx,
            wait_for_completion=wait_for_completion,
            node_id=self.node_id,
            parameter=intensity,
        )
        await command.send()
        await self.after_update()

    async def turn_on(self, wait_for_completion: bool = True) -> None:
        """Turn on light.

        Parameters:
            * wait_for_completion: If set, function will return
                after device has reached target intensity.

        """
        await self.set_intensity(
            intensity=Intensity(intensity_percent=100),
            wait_for_completion=wait_for_completion,
        )

    async def turn_off(self, wait_for_completion: bool = True) -> None:
        """Turn off light.

        Parameters:
            * wait_for_completion: If set, function will return
                after device has reached target intensity.

        """
        await self.set_intensity(
            intensity=Intensity(intensity_percent=0),
            wait_for_completion=wait_for_completion,
        )

    def __str__(self) -> str:
        """Return object as readable string."""
        return (
            '<{} name="{}" '
            'node_id="{}" '
            'serial_number="{}"/>'.format(
                type(self).__name__, self.name, self.node_id, self.serial_number
            )
        )


class ExteriorHeating(DimmableDevice):
    """Exterior heating device that supports setting intensity."""


class Light(DimmableDevice):
    """Lights that support setting brightness."""


class OnOffLight(DimmableDevice):
    """Lights supporting on/off only."""



"""Module for retrieving scene list from API."""
from typing import TYPE_CHECKING, List, Optional, Tuple

from pyvlx.log import PYVLXLOG

from .api_event import ApiEvent
from .frames import (
    FrameBase, FrameGetSceneListConfirmation, FrameGetSceneListNotification,
    FrameGetSceneListRequest)

if TYPE_CHECKING:
    from pyvlx import PyVLX


class GetSceneList(ApiEvent):
    """Class for retrieving scene list from API."""

    def __init__(self, pyvlx: "PyVLX"):
        """Initialize SceneList class."""
        super().__init__(pyvlx=pyvlx)
        self.success = False
        self.count_scenes: Optional[int] = None
        self.scenes: List[Tuple[int, str]] = []

    async def handle_frame(self, frame: FrameBase) -> bool:
        """Handle incoming API frame, return True if this was the expected frame."""
        if isinstance(frame, FrameGetSceneListConfirmation):
            self.count_scenes = frame.count_scenes
            if self.count_scenes == 0:
                self.success = True
                return True
            # We are still waiting for FrameGetSceneListNotification(s)
            return False
        if isinstance(frame, FrameGetSceneListNotification):
            self.scenes.extend(frame.scenes)
            if frame.remaining_scenes != 0:
                # We are still waiting for FrameGetSceneListConfirmation(s)
                return False
            if self.count_scenes != len(self.scenes):
                PYVLXLOG.warning(
                    "Warning: number of received scenes does not match expected number"
                )
            self.success = True
            return True
        return False

    def request_frame(self) -> FrameGetSceneListRequest:
        """Construct initiating frame."""
        return FrameGetSceneListRequest()



"""Unit tests for FrameCommandRunStatusNotification."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameCommandRunStatusNotification
from pyvlx.const import RunStatus, StatusReply


class TestFrameCommandRunStatusNotification(unittest.TestCase):
    """Test class FrameCommandRunStatusNotification."""

    # pylint: disable=too-many-public-methods,invalid-name

    EXAMPLE_FRAME = (
        b"\x00\x10\x03\x02\x03\xe8\x07\x17*\x059\x01\xe3\x00\x00\x00\x00\x1e"
    )

    def test_bytes(self):
        """Test FrameCommandRunStatusNotification."""
        frame = FrameCommandRunStatusNotification(
            session_id=1000,
            status_id=7,
            index_id=23,
            node_parameter=42,
            parameter_value=1337,
            run_status=RunStatus.EXECUTION_FAILED,
            status_reply=StatusReply.LIMITATION_BY_RAIN,
        )
        self.assertEqual(bytes(frame), self.EXAMPLE_FRAME)

    def test_frame_from_raw(self):
        """Test parse FrameCommandRunStatusNotification from raw."""
        frame = frame_from_raw(self.EXAMPLE_FRAME)
        self.assertIsInstance(frame, FrameCommandRunStatusNotification)
        self.assertEqual(frame.session_id, 1000)
        self.assertEqual(frame.status_id, 7)
        self.assertEqual(frame.index_id, 23)
        self.assertEqual(frame.node_parameter, 42)
        self.assertEqual(frame.parameter_value, 1337)
        self.assertEqual(frame.run_status, RunStatus.EXECUTION_FAILED)
        self.assertEqual(frame.status_reply, StatusReply.LIMITATION_BY_RAIN)

    def test_str(self):
        """Test string representation of FrameCommandRunStatusNotification."""
        frame = FrameCommandRunStatusNotification(
            session_id=1000,
            status_id=7,
            index_id=23,
            node_parameter=42,
            parameter_value=1337,
            run_status=RunStatus.EXECUTION_FAILED,
            status_reply=StatusReply.LIMITATION_BY_RAIN,
        )
        self.assertEqual(
            str(frame),
            '<FrameCommandRunStatusNotification session_id="1000" status_id="7" index_id="23" node_parameter="42" '
            'parameter_value="1337" run_status="RunStatus.EXECUTION_FAILED" '
            'status_reply="StatusReply.LIMITATION_BY_RAIN"/>',
        )



"""Unit tests for FrameGetNetworkSetupConfirmation."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import DHCPParameter, FrameGetNetworkSetupConfirmation


class TestFrameGetNetworkSetupConfirmation(unittest.TestCase):
    """Test class for FrameGetNetworkSetupConfirmation."""

    # pylint: disable=too-many-public-methods,invalid-name
    TESTFRAME = bytes.fromhex('001000e1c0a80de3ffffff00c0a80d0100ec')

    def test_bytes(self):
        """Test FrameGetNetworkSetupConfirmation."""
        frame = FrameGetNetworkSetupConfirmation(
            ipaddress=b'\xc0\xa8\r\xe3', netmask=b'\xff\xff\xff\x00',
            gateway=b'\xc0\xa8\r\x01', dhcp=DHCPParameter.ENABLE)
        self.assertEqual(bytes(frame),
                         b"\x00\x10\x00\xe1\xc0\xa8\r\xe3\xff\xff\xff\x00\xc0\xa8\r\x01\x00\xec")

    def test_frame_from_raw(self):
        """Test parse FrameGetNetworkSetupConfirmation from raw."""
        frame = frame_from_raw(self.TESTFRAME)
        self.assertTrue(isinstance(frame, FrameGetNetworkSetupConfirmation))
        self.assertEqual(frame.ipaddress, '192.168.13.227')
        self.assertEqual(frame.netmask, '255.255.255.0')
        self.assertEqual(frame.gateway, '192.168.13.1')
        self.assertEqual(frame.dhcp, DHCPParameter.DISABLE)

    def test_str(self):
        """Test string representation of FrameGetNetworkSetupConfirmation."""
        frame = FrameGetNetworkSetupConfirmation(
            ipaddress=b'\xc0\xa8\r\xe3', netmask=b'\xff\xff\xff\x00',
            gateway=b'\xc0\xa8\r\x01', dhcp=DHCPParameter.DISABLE)
        self.assertEqual(
            str(frame),
            '<FrameGetNetworkSetupConfirmation ipaddress="192.168.13.227" '
            'netmask="255.255.255.0" gateway="192.168.13.1" dhcp="DHCPParameter.DISABLE"/>',
        )



"""Module for get node information from gateway."""
from enum import Enum
from typing import Dict, List, Optional

from pyvlx.const import (
    Command, NodeParameter, RunStatus, StatusReply, StatusType)
from pyvlx.exception import PyVLXException
from pyvlx.parameter import Parameter

from .frame import FrameBase


class FrameStatusRequestRequest(FrameBase):
    """Frame for status request request."""

    PAYLOAD_LEN = 26

    def __init__(self, session_id: Optional[int] = None, node_ids: Optional[List[int]] = None):
        """Init Frame."""
        super().__init__(Command.GW_STATUS_REQUEST_REQ)
        self.session_id = session_id
        self.node_ids = node_ids if node_ids is not None else []
        self.status_type = StatusType.REQUEST_CURRENT_POSITION
        self.fpi1 = 254     # Request FP1 to FP7
        self.fpi2 = 0

    def get_payload(self) -> bytes:
        """Return Payload."""
        assert self.session_id is not None
        ret = bytes([self.session_id >> 8 & 255, self.session_id & 255])
        ret += bytes([len(self.node_ids)])      # index array count
        ret += bytes(self.node_ids) + bytes(20 - len(self.node_ids))
        ret += bytes([self.status_type.value])
        ret += bytes([self.fpi1])
        ret += bytes([self.fpi2])
        return ret

    def from_payload(self, payload: bytes) -> None:
        """Init frame from binary data."""
        self.session_id = payload[0] * 256 + payload[1]
        len_node_ids = payload[2]
        if len_node_ids > 20:
            raise PyVLXException("command_send_request_wrong_node_length")
        self.node_ids = []
        for i in range(len_node_ids):
            self.node_ids.append(payload[3] + i)

        self.status_type = StatusType(payload[23])
        self.fpi1 = payload[24]
        self.fpi2 = payload[25]

    def __str__(self) -> str:
        """Return human readable string."""
        return (
            '<{} session_id="{}" node_ids="{}" '
            'status_type="{}" fpi1="{}" fpi2="{}"/>'.format(
                type(self).__name__, self.session_id,
                self.node_ids,
                self.status_type, self.fpi1, self.fpi2
            )
        )


class StatusRequestStatus(Enum):
    """Enum for status request status."""

    REJECTED = 0
    ACCEPTED = 1


class FrameStatusRequestConfirmation(FrameBase):
    """Frame for confirmation for status request request."""

    PAYLOAD_LEN = 3

    def __init__(self, session_id: Optional[int] = None, status: Optional[StatusRequestStatus] = None):
        """Init Frame."""
        super().__init__(Command.GW_STATUS_REQUEST_CFM)
        self.session_id = session_id
        self.status = status

    def get_payload(self) -> bytes:
        """Return Payload."""
        assert self.session_id is not None
        ret = bytes([self.session_id >> 8 & 255, self.session_id & 255])
        assert self.status is not None
        ret += bytes([self.status.value])
        return ret

    def from_payload(self, payload: bytes) -> None:
        """Init frame from binary data."""
        self.session_id = payload[0] * 256 + payload[1]
        self.status = StatusRequestStatus(payload[2])

    def __str__(self) -> str:
        """Return human readable string."""
        return '<{} session_id="{}" status="{}"/>'.format(
            type(self).__name__, self.session_id, self.status
        )


class FrameStatusRequestNotification(FrameBase):
    """Frame for notification of status request request."""

    # PAYLOAD_LEN = 59
    # No PAYLOAD_LEN because it is variable depending on StatusType

    def __init__(self) -> None:
        """Init Frame."""
        super().__init__(Command.GW_STATUS_REQUEST_NTF)
        self.session_id = 0
        self.status_id = 0
        self.node_id = 0
        self.run_status = RunStatus.EXECUTION_COMPLETED
        self.status_reply = StatusReply.UNKNOWN_STATUS_REPLY
        self.status_type = StatusType.REQUEST_TARGET_POSITION
        self.status_count = 0
        self.parameter_data: Dict[NodeParameter, Parameter] = {}
        self.target_position = Parameter()
        self.current_position = Parameter()
        self.remaining_time = 0
        self.last_master_execution_address = b''
        self.last_command_originator = 0

    def get_payload(self) -> bytes:
        """Return Payload."""
        payload = bytes()
        payload += bytes([self.session_id >> 8 & 255, self.session_id & 255])
        payload += bytes([self.status_id])
        payload += bytes([self.node_id])
        payload += bytes([self.run_status.value])
        payload += bytes([self.status_reply.value])
        payload += bytes([self.status_type.value])
        if self.status_type == StatusType.REQUEST_MAIN_INFO:
            payload += bytes(self.target_position.raw)
            payload += bytes(self.current_position.raw)
            payload += bytes([self.remaining_time >> 8 & 255, self.remaining_time & 255])
            payload += self.last_master_execution_address
            payload += bytes([self.last_command_originator])
        else:
            payload += bytes([self.status_count])
            keys = self.parameter_data.keys()
            for key in keys:
                payload += bytes([key.value])
                payload += bytes(self.parameter_data[key].raw)
            payload += bytes(51 - len(self.parameter_data))

        return payload

    def from_payload(self, payload: bytes) -> None:
        """Init frame from binary data."""
        self.session_id = payload[0] * 256 + payload[1]
        self.status_id = payload[2]
        self.node_id = payload[3]
        self.run_status = RunStatus(payload[4])
        self.status_reply = StatusReply(payload[5])
        self.status_type = StatusType(payload[6])
        if self.status_type == StatusType.REQUEST_MAIN_INFO:
            self.target_position = Parameter(payload[7:8])
            self.current_position = Parameter(payload[9:10])
            self.remaining_time = payload[11] * 256 + payload[12]
            self.last_master_execution_address = payload[13:16]
            self.last_command_originator = payload[17]
        else:
            self.status_count = payload[7]
            for i in range(8, 8 + self.status_count*3, 3):
                self.parameter_data.update({NodeParameter(payload[i]): Parameter(payload[i+1:i+3])})

    def __str__(self) -> str:
        """Return human readable string."""
        if self.status_type == StatusType.REQUEST_MAIN_INFO:
            return (
                '<{} session_id="{}" status_id="{}" '
                'node_id="{}" run_status="{}" status_reply="{}" status_type="{}" target_position="{}" '
                'current_position="{}" remaining_time="{}" last_master_execution_address="{!r}" last_command_originator="{}"/>'.format(
                    type(self).__name__,
                    self.session_id,
                    self.status_id,
                    self.node_id,
                    self.run_status,
                    self.status_reply,
                    self.status_type,
                    self.target_position,
                    self.current_position,
                    self.remaining_time,
                    self.last_master_execution_address,
                    self.last_command_originator,
                )
            )

        parameter_data_str = ""
        for key, value in self.parameter_data.items():
            parameter_data_str += "%s: %s, " % (
                str(key),
                str(value),
            )

        return (
            '<{} session_id="{}" status_id="{}" '
            'node_id="{}" run_status="{}" status_reply="{}" status_type="{}" status_count="{}" '
            'parameter_data="{}"/>'.format(
                type(self).__name__,
                self.session_id,
                self.status_id,
                self.node_id,
                self.run_status,
                self.status_reply,
                self.status_type,
                self.status_count,
                parameter_data_str
            )
        )



"""Just a demo of the new PyVLX module."""
import asyncio
import logging

from pyvlx import PyVLX


async def main(loop):
    """Log packets from Bus."""
    # Setting debug
    logging.basicConfig(level=logging.DEBUG)

    # Connecting to KLF 200
    pyvlx = PyVLX('pyvlx.yaml', loop=loop)
    await pyvlx.load_scenes()
    await pyvlx.load_nodes()

    # and wait, increase this timeout if you want to
    # log for a longer time.:)
    await asyncio.sleep(90)

    # Cleanup, KLF 200 is terrible in handling lost connections
    await pyvlx.disconnect()


if __name__ == '__main__':
    myloop = asyncio.new_event_loop()
    myloop.run_until_complete(main(myloop))
    myloop.close()



"""Unit tests for FrameGetAllNodesInformationConfirmation."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameGetAllNodesInformationConfirmation


class TestFrameGetAllNodesInformationConfirmation(unittest.TestCase):
    """Test class for FrameGetAllNodesInformationConfirmation."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_bytes(self):
        """Test FrameGetAllNodesInformationConfirmation."""
        frame = FrameGetAllNodesInformationConfirmation(number_of_nodes=23)
        self.assertEqual(bytes(frame), b"\x00\x05\x02\x03\x00\x17\x13")

    def test_frame_from_raw(self):
        """Test parse FrameGetAllNodesInformationConfirmation from raw."""
        frame = frame_from_raw(b"\x00\x05\x02\x03\x00\x17\x13")
        self.assertTrue(isinstance(frame, FrameGetAllNodesInformationConfirmation))
        self.assertEqual(frame.number_of_nodes, 23)

    def test_str(self):
        """Test string representation of FrameGetAllNodesInformationConfirmation."""
        frame = FrameGetAllNodesInformationConfirmation(number_of_nodes=23)
        self.assertEqual(
            str(frame),
            '<FrameGetAllNodesInformationConfirmation status="AllNodesInformationStatus.OK" number_of_nodes="23"/>',
        )



"""Unit test for command send module."""
import unittest
from unittest.mock import AsyncMock, MagicMock, patch

from pyvlx import Parameter, PyVLXException
from pyvlx.api import CommandSend
from pyvlx.api.frames import (
    CommandSendConfirmationStatus, FrameCommandRemainingTimeNotification,
    FrameCommandRunStatusNotification, FrameCommandSendConfirmation,
    FrameCommandSendRequest, FrameSessionFinishedNotification)


class TestCommandSend(unittest.IsolatedAsyncioTestCase):
    """Test class for CommandSend."""

    def setUp(self) -> None:
        """Set up TestCommandSend."""
        mocked_pyvlx = MagicMock()
        self.command_send = CommandSend(pyvlx=mocked_pyvlx, node_id=23, parameter=Parameter())

    async def test_handle_frame(self) -> None:
        """Test handle_frame function of CommandSend object."""
        frame = MagicMock(spec=FrameCommandSendConfirmation)
        session_id = 1
        self.command_send.session_id = session_id
        frame.session_id = session_id
        self.command_send.wait_for_completion = False
        frame.status = CommandSendConfirmationStatus.ACCEPTED
        self.assertTrue(await self.command_send.handle_frame(frame=frame))
        self.assertTrue(self.command_send.success)

        self.command_send.success = False
        self.command_send.wait_for_completion = True
        frame.status = CommandSendConfirmationStatus.ACCEPTED
        self.assertFalse(await self.command_send.handle_frame(frame=frame))
        self.assertTrue(self.command_send.success)

        self.command_send.success = False
        self.command_send.wait_for_completion = False
        frame.status = CommandSendConfirmationStatus.REJECTED
        self.assertTrue(await self.command_send.handle_frame(frame=frame))
        self.assertFalse(self.command_send.success)

        self.command_send.success = False
        self.command_send.wait_for_completion = True
        frame.status = CommandSendConfirmationStatus.REJECTED
        self.assertFalse(await self.command_send.handle_frame(frame=frame))
        self.assertFalse(self.command_send.success)

        frame = MagicMock(spec=FrameCommandRemainingTimeNotification)
        frame.session_id = session_id
        self.assertFalse(await self.command_send.handle_frame(frame=frame))

        frame = MagicMock(spec=FrameCommandRunStatusNotification)
        frame.session_id = session_id
        self.assertFalse(await self.command_send.handle_frame(frame=frame))

        frame = MagicMock(spec=FrameSessionFinishedNotification)
        frame.session_id = session_id
        self.assertTrue(await self.command_send.handle_frame(frame=frame))

        frame = MagicMock(spec=FrameSessionFinishedNotification)
        frame.session_id = session_id + 1
        self.assertFalse(await self.command_send.handle_frame(frame=frame))

    @patch("pyvlx.api.ApiEvent.do_api_call", new_callable=AsyncMock)
    async def test_send(self, do_api_call: AsyncMock) -> None:
        """Test send function of CommandSend object."""
        self.command_send.success = True
        await self.command_send.send()
        assert do_api_call.called

        self.command_send.success = False
        with self.assertRaises(PyVLXException):
            await self.command_send.send()

    @patch("pyvlx.api.command_send.get_new_session_id", callable=MagicMock)
    def test_request_frame(self, new_session_id_request: MagicMock) -> None:
        """Test request_frame function of CommandSend object."""
        self.command_send.session_id = 1
        new_session_id_request.return_value = 5
        assert isinstance(self.command_send.request_frame(), FrameCommandSendRequest)
        assert new_session_id_request.called
        assert self.command_send.session_id == 5



"""Unit tests for FrameGetNodeInformationRequest."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameGetNodeInformationRequest


class TestFrameGetNodeInformationRequest(unittest.TestCase):
    """Test class for FrameGetNodeInformationRequest."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_bytes(self):
        """Test FrameGetNodeInformationRequest with NO_TYPE."""
        frame = FrameGetNodeInformationRequest(node_id=23)
        self.assertEqual(bytes(frame), b"\x00\x04\x02\x00\x17\x11")

    def test_frame_from_raw(self):
        """Test parse FrameGetNodeInformationRequest from raw."""
        frame = frame_from_raw(b"\x00\x04\x02\x00\x17\x11")
        self.assertTrue(isinstance(frame, FrameGetNodeInformationRequest))
        self.assertEqual(frame.node_id, 23)

    def test_str(self):
        """Test string representation of FrameGetNodeInformationRequest."""
        frame = FrameGetNodeInformationRequest(node_id=23)
        self.assertEqual(str(frame), '<FrameGetNodeInformationRequest node_id="23"/>')



"""Unit tests for FrameCommandSendRequest."""
import unittest

from pyvlx import Position, PyVLXException
from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameCommandSendRequest
from pyvlx.const import Originator


class TestFrameCommandSendRequest(unittest.TestCase):
    """Test class FrameCommandSendRequest."""

    # pylint: disable=too-many-public-methods,invalid-name

    EXAMPLE_FRAME = (
        b"\x00E\x03\x00\x03\xe8\x02\x03\x00\x00\x00\x96\x00\x00\x00\x00"
        + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x03\x01\x02"
        + b"\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        + b"\x00\x00\x00\x00\x00\x009"
    )

    def test_bytes(self) -> None:
        """Test FrameCommandSendRequest with NO_TYPE."""
        frame = FrameCommandSendRequest(
            node_ids=[1, 2, 3],
            parameter=Position(position_percent=75),
            session_id=1000,
            originator=Originator.RAIN,
        )
        self.assertEqual(bytes(frame), self.EXAMPLE_FRAME)

    def test_frame_from_raw(self) -> None:
        """Test parse FrameCommandSendRequest from raw."""
        frame = frame_from_raw(self.EXAMPLE_FRAME)
        self.assertTrue(isinstance(frame, FrameCommandSendRequest))
        self.assertEqual(frame.node_ids, [1, 2, 3])
        self.assertEqual(Position(frame.parameter).position_percent, 75)
        self.assertEqual(frame.session_id, 1000)
        self.assertEqual(frame.originator, Originator.RAIN)

    def test_str(self) -> None:
        """Test string representation of FrameCommandSendRequest."""
        functional_para = {"fp3": Position(position=12345)}
        functional_parameter = {}
        for i in range(1, 17):
            key = "fp%s" % (i)
            if key in functional_para:
                functional_parameter[key] = functional_para[key]
            else:
                functional_parameter[key] = bytes(2)
        frame = FrameCommandSendRequest(
            node_ids=[1, 2, 3],
            parameter=Position(position=12345),
            **functional_parameter,
            session_id=1000,
            originator=Originator.RAIN
        )
        self.assertEqual(
            str(frame),
            '<FrameCommandSendRequest node_ids="[1, 2, 3]" active_parameter="0" parameter="24 %" '
            'functional_parameter="fp1: 0 %, fp2: 0 %, fp3: 24 %, fp4: 0 %, fp5: 0 %, fp6: 0 %, fp7: 0 %, '
            'fp8: 0 %, fp9: 0 %, fp10: 0 %, fp11: 0 %, fp12: 0 %, fp13: 0 %, fp14: 0 %, fp15: 0 %, fp16: 0 %, " '
            'session_id="1000" originator="Originator.RAIN"/>',
        )

    def test_wrong_payload(self) -> None:
        """Test wrong payload length, 2 scenes in len, only one provided."""
        frame = FrameCommandSendRequest()
        with self.assertRaises(PyVLXException) as ctx:
            frame.from_payload(
                b"\x03\xe8\x01\x03\x00\x00\x0009\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x15\x01\x02\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00"
            )
        self.assertEqual(
            ctx.exception.description, "command_send_request_wrong_node_length"
        )



"""Unit test for scene."""

import unittest
from unittest.mock import MagicMock

from pyvlx import Scene


# pylint: disable=too-many-public-methods,invalid-name
class TestScene(unittest.TestCase):
    """Test class for scene."""

    def test_get_name(self) -> None:
        """Test get_name()."""
        pyvlx = MagicMock()
        scene = Scene(pyvlx, 2, "Scene 1")
        self.assertEqual(scene.name, "Scene 1")
        self.assertEqual(scene.scene_id, 2)

    def test_str(self) -> None:
        """Test string representation of Scene object."""
        pyvlx = MagicMock()
        scene = Scene(pyvlx, 2, "Scene 1")
        self.assertEqual(str(scene), '<Scene name="Scene 1" id="2"/>')



"""Unit tests for PyVLX PasswordEnterRequest."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FramePasswordEnterRequest
from pyvlx.exception import PyVLXException


class TestFramePasswordEnter(unittest.TestCase):
    """Test class for FramePasswordEnterRequest."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_bytes(self):
        """Test FramePasswordEnterRequest."""
        frame = FramePasswordEnterRequest(password="fnord")
        self.assertEqual(
            bytes(frame),
            b"\x00#0\x00fnord\x00\x00\x00\x00\x00"
            + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00b",
        )

    def test_bytes_long_pw(self):
        """Test FramePasswordEnterRequest with long password."""
        frame = FramePasswordEnterRequest(password="x" * 32)
        self.assertEqual(
            bytes(frame), b"\x00#0\x00xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx\x13"
        )

    def test_frame_from_raw(self):
        """Test parsing FramePasswordEnterRequest from raw bytes."""
        frame = frame_from_raw(
            b"\x00#0\x00fnord\x00\x00\x00\x00\x00"
            + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00b"
        )
        self.assertTrue(isinstance(frame, FramePasswordEnterRequest))
        self.assertEqual(frame.password, "fnord")

    def test_frame_from_raw_long(self):
        """Test parsing FramePasswordEnterRequest from raw bytes with long password."""
        frame = frame_from_raw(b"\x00#0\x00xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx\x13")
        self.assertTrue(isinstance(frame, FramePasswordEnterRequest))
        self.assertEqual(frame.password, "x" * 32)

    def test_errors(self):
        """Test FramePasswordEnterRequest with wrong password."""
        with self.assertRaises(PyVLXException):
            bytes(FramePasswordEnterRequest())
        with self.assertRaises(PyVLXException):
            bytes(FramePasswordEnterRequest(password="x" * 33))

    def test_str(self):
        """Test string representation of FramePasswordEnterRequest."""
        frame = FramePasswordEnterRequest(password="fnord")
        self.assertEqual(str(frame), '<FramePasswordEnterRequest password="fn****"/>')

    def test_str_no_password(self):
        """Test string representation of FramePasswordEnterRequest with no password."""
        frame = FramePasswordEnterRequest()
        self.assertEqual(str(frame), '<FramePasswordEnterRequest password="None"/>')



"""Unit tests for FrameGetLocalTimeConfirmation."""
import unittest
from datetime import datetime, timezone

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameGetLocalTimeConfirmation


class TestFrameGetLocalTimeConfirmation(unittest.TestCase):
    """Test class for FrameGetLocalTimeConfirmation."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_bytes(self) -> None:
        """Test FrameGetLocalTimeConfirmation."""
        frame = FrameGetLocalTimeConfirmation()
        frame.time.localtime = datetime(2020, 12, 3, 18, 19, 19, 176900)
        frame.time.utctime = datetime(
            2020, 12, 3, 18, 19, 19, 176900, tzinfo=timezone.utc
        )
        self.assertEqual(
            bytes(frame), b"\x00\x12 \x05_\xc9,'\x13\x13\x12\x03\x0c\x00x\x04\x01R\xffg"
        )

    def test_frame_from_raw(self) -> None:
        """Test parse FrameGetLocalTimeConfirmation from raw."""
        frame = frame_from_raw(
            b"\x00\x12 \x05_\xc9,'\x13\x13\x12\x03\x0c\x00x\x04\x01R\xffg"
        )
        self.assertTrue(isinstance(frame, FrameGetLocalTimeConfirmation))

    def test_str(self) -> None:
        """Test string representation of FrameGetLocalTimeConfirmation."""
        frame = FrameGetLocalTimeConfirmation()
        frame.time.localtime = datetime.strptime(
            "2020-12-03 18:19:19.176900", "%Y-%m-%d %H:%M:%S.%f"
        )
        frame.time.utctime = datetime.strptime(
            "2020-12-03 18:19:19.176900", "%Y-%m-%d %H:%M:%S.%f"
        )
        self.assertEqual(
            str(frame),
            "<FrameGetLocalTimeConfirmation><DtoLocalTime "
            'utctime="2020-12-03 18:19:19.176900" '
            'localtime="2020-12-03 18:19:19.176900"/>'
            "</FrameGetLocalTimeConfirmation>",
        )



"""Unit test for limitation."""
import asyncio
import unittest
from unittest.mock import MagicMock

import pytest
from pytest import FixtureRequest

from pyvlx import PyVLX
from pyvlx.api.frames.frame_get_limitation import (
    FrameGetLimitationStatus, FrameGetLimitationStatusConfirmation,
    FrameGetLimitationStatusNotification)
from pyvlx.api.get_limitation import GetLimitation
from pyvlx.const import LimitationType, Originator


@pytest.fixture(scope="class")
def event_loop_instance(request: FixtureRequest) -> None:
    """Add the event_loop as an attribute to the unittest style test class."""
    request.cls.event_loop = asyncio.get_event_loop_policy().new_event_loop()
    yield
    request.cls.event_loop.close()


# pylint: disable=too-many-public-methods,invalid-name
@pytest.mark.usefixtures("event_loop_instance")
class TestGetLimitation(unittest.TestCase):
    """Test class for Limitation."""

    def setUp(self) -> None:
        """Set up TestGetLimitation."""
        self.pyvlx = MagicMock(spec=PyVLX)

    def test_get_name(self) -> None:
        """Test get_name()."""
        limit = GetLimitation(self.pyvlx, 1)
        self.assertEqual(limit.node_id, 1)
        self.assertEqual(limit.limitation_type, LimitationType.MIN_LIMITATION)
        limit = GetLimitation(self.pyvlx, 2, LimitationType.MAX_LIMITATION)
        self.assertEqual(limit.node_id, 2)
        self.assertEqual(limit.limitation_type, LimitationType.MAX_LIMITATION)

    def test_max_value(self) -> None:
        """Test limit.max_value."""
        limit = GetLimitation(self.pyvlx, 1)
        limit.max_value_raw = b'\xf7'
        self.assertEqual(limit.max_value, 124)

    def test_min_value(self) -> None:
        """Test limit.min_value."""
        limit = GetLimitation(self.pyvlx, 1)
        limit.min_value_raw = b'\xba'
        self.assertEqual(limit.min_value, 93)

    def test_handle_frame(self) -> None:
        """Test handle frame."""
        limit = GetLimitation(self.pyvlx, 1)

        frame = FrameGetLimitationStatus()
        self.assertFalse(self.event_loop.run_until_complete(limit.handle_frame(frame)))
        self.assertFalse(limit.success)

        frame = FrameGetLimitationStatusConfirmation()
        self.assertFalse(self.event_loop.run_until_complete(limit.handle_frame(frame)))
        self.assertFalse(limit.success)

        frame = FrameGetLimitationStatusNotification()
        frame.session_id = 1
        frame.node_id = 1
        frame.min_value = b'\xf7'
        frame.max_value = b'\xba'
        frame.limit_originator = Originator.USER
        frame.limit_time = 1

        limit.session_id = 0
        self.assertFalse(self.event_loop.run_until_complete(limit.handle_frame(frame)))
        self.assertFalse(limit.success)  # Session id is wrong

        limit.session_id = frame.session_id
        self.assertTrue(self.event_loop.run_until_complete(limit.handle_frame(frame)))
        self.assertTrue(limit.success)
        self.assertEqual(limit.node_id, frame.node_id)
        self.assertEqual(limit.session_id, frame.session_id)
        self.assertEqual(limit.min_value, 124)
        self.assertEqual(limit.max_value, 93)
        self.assertEqual(limit.originator, frame.limit_originator)
        self.assertEqual(limit.limit_time, frame.limit_time)

    def test_request_frame(self) -> None:
        """Test initiating frame."""
        limit = GetLimitation(self.pyvlx, 1)
        req_frame = limit.request_frame()
        self.assertIsInstance(req_frame, FrameGetLimitationStatus)
        self.assertTrue(req_frame.session_id, 1)
        self.assertTrue(req_frame.node_ids, [1])
        self.assertTrue(req_frame.limitations_type, limit.limitation_type)

        limit.limitation_type = LimitationType.MAX_LIMITATION
        self.assertIsInstance(req_frame, FrameGetLimitationStatus)
        self.assertTrue(req_frame.session_id, 1)
        self.assertTrue(req_frame.limitations_type, limit.limitation_type)



"""Unit tests for FrameGetSceneListConfirmation."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameGetSceneListConfirmation


class TestFrameGetSceneListConfirmation(unittest.TestCase):
    """Test class for FrameGetSceneListConfirmation."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_bytes(self):
        """Test FrameGetSceneListConfirmation."""
        frame = FrameGetSceneListConfirmation(count_scenes=12)
        self.assertEqual(bytes(frame), b"\x00\x04\x04\r\x0c\x01")

    def test_frame_from_raw(self):
        """Test parse FrameGetSceneListConfirmation from raw."""
        frame = frame_from_raw(b"\x00\x04\x04\r\x0c\x01")
        self.assertTrue(isinstance(frame, FrameGetSceneListConfirmation))
        self.assertEqual(frame.count_scenes, 12)

    def test_str(self):
        """Test string representation of FrameGetSceneListConfirmation."""
        frame = FrameGetSceneListConfirmation(count_scenes=12)
        self.assertEqual(str(frame), '<FrameGetSceneListConfirmation count_scenes="12"/>')




"""Module for get local time classes."""
from typing import List, Optional

from pyvlx.const import Command, LimitationType, Originator, Priority

from .frame import FrameBase


class FrameGetLimitationStatus(FrameBase):
    """Frame for requesting limitation status."""

    PAYLOAD_LEN = 25

    def __init__(self,
                 node_ids: Optional[List[int]] = None,
                 session_id: Optional[int] = None,
                 limitation_type: LimitationType = LimitationType.MIN_LIMITATION):
        """Init Frame."""
        super().__init__(Command.GW_GET_LIMITATION_STATUS_REQ)
        self.session_id = session_id
        self.originator = Originator.USER
        self.priority = Priority.USER_LEVEL_2
        self.node_ids = node_ids if node_ids is not None else []

        self.parameter_id = 0  # Main Parameter
        self.limitations_type = limitation_type

    def get_payload(self) -> bytes:
        """Return Payload."""
        assert self.session_id is not None
        ret = bytes([self.session_id >> 8 & 255, self.session_id & 255])
        ret += bytes([len(self.node_ids)])  # index array count
        ret += bytes(self.node_ids) + bytes(20 - len(self.node_ids))
        ret += bytes([self.parameter_id])
        ret += bytes([self.limitations_type.value])
        return ret

    def __str__(self) -> str:
        """Return human readable string."""
        return f'<{type(self).__name__} node_ids="{self.node_ids}" ' \
               f'session_id="{self.session_id}" originator="{self.originator}" />'


class FrameGetLimitationStatusConfirmation(FrameBase):
    """Frame for response for get limitation requests."""

    PAYLOAD_LEN = 3

    def __init__(self, session_id: Optional[int] = None, data: Optional[int] = None):
        """Init Frame."""
        super().__init__(Command.GW_GET_LIMITATION_STATUS_CFM)
        self.session_id = session_id
        self.data = data

    def get_payload(self) -> bytes:
        """Return Payload."""
        assert self.session_id is not None
        ret = bytes([self.session_id >> 8 & 255, self.session_id & 255])
        assert self.data is not None
        ret += bytes([self.data])
        return ret

    def from_payload(self, payload: bytes) -> None:
        """Init frame from binary data."""
        self.session_id = payload[0] * 256 + payload[1]
        self.data = payload[2]

    def __str__(self) -> str:
        """Return human readable string."""
        return '<{} session_id="{}" status="{}"/>'.format(
            type(self).__name__, self.session_id, self.data
        )


class FrameGetLimitationStatusNotification(FrameBase):
    """Frame for notification of note information request."""

    PAYLOAD_LEN = 10

    def __init__(self) -> None:
        """Init Frame."""
        super().__init__(Command.GW_LIMITATION_STATUS_NTF)
        self.session_id: Optional[int] = None
        self.node_id = 0
        self.parameter_id = 0
        self.min_value: Optional[bytes] = None
        self.max_value: Optional[bytes] = None
        self.limit_originator: Optional[Originator] = None
        self.limit_time: Optional[int] = None

    def get_payload(self) -> bytes:
        """Return Payload."""
        assert self.session_id is not None
        assert self.min_value is not None
        assert self.max_value is not None
        assert self.limit_originator is not None
        assert self.limit_time is not None
        payload = bytes([self.session_id >> 8 & 255, self.session_id & 255])
        payload += bytes([self.node_id])
        payload += bytes([self.parameter_id])
        payload += self.min_value
        payload += self.max_value
        payload += bytes([self.limit_originator.value])
        payload += bytes([self.limit_time])
        return payload

    def from_payload(self, payload: bytes) -> None:
        """Init frame from binary data."""
        self.session_id = payload[0] * 256 + payload[1]
        self.node_id = payload[2]
        self.parameter_id = payload[3]
        self.min_value = payload[4:5]
        self.max_value = payload[6:7]
        self.limit_originator = Originator(payload[8])
        self.limit_time = payload[9]

    def __str__(self) -> str:
        """Return human readable string."""
        return (
            '<{} node_id="{}" session_id="{}" min_value="{!r}" '
            'max_value="{!r}" originator="{}" limit_time="{}"/>'.format(
                type(self).__name__, self.node_id, self.session_id,
                self.min_value, self.max_value, self.limit_originator,
                self.limit_time
            )
        )



"""Unit tests for FrameGetAllNodesInformationFinishedNotification."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameGetAllNodesInformationFinishedNotification


class TestFrameGetAllNodesInformationFinishedNotification(unittest.TestCase):
    """Test class for FrameGetAllNodesInformationFinishedNotification."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_bytes(self):
        """Test FrameGetAllNodesInformationFinishedNotification."""
        frame = FrameGetAllNodesInformationFinishedNotification()
        self.assertEqual(bytes(frame), b"\x00\x03\x02\x05\x04")

    def test_frame_from_raw(self):
        """Test parse FrameGetAllNodesInformationFinishedNotification from raw."""
        frame = frame_from_raw(b"\x00\x03\x02\x05\x04")
        self.assertTrue(
            isinstance(frame, FrameGetAllNodesInformationFinishedNotification)
        )

    def test_str(self):
        """Test string representation of FrameGetAllNodesInformationFinishedNotification."""
        frame = FrameGetAllNodesInformationFinishedNotification()
        self.assertEqual(
            str(frame), "<FrameGetAllNodesInformationFinishedNotification/>"
        )



"""Unit test for configuration."""
import unittest
from asyncio import AbstractEventLoop
from unittest.mock import MagicMock

from pyvlx import PyVLX, PyVLXException


# pylint: disable=too-many-public-methods,invalid-name
class TestConfig(unittest.TestCase):
    """Test class for configuration."""

    def setUp(self) -> None:
        """Set up TestCommandSend."""
        self.mocked_loop = MagicMock(spec=AbstractEventLoop)

    def test_config_file(self) -> None:
        """Test host/password configuration via config.yaml."""
        pyvlx = PyVLX(loop=self.mocked_loop, path="test/data/test_config.yaml")  # pylint: disable=E1123
        self.assertEqual(pyvlx.config.password, "velux123")
        self.assertEqual(pyvlx.config.host, "192.168.2.127")
        self.assertEqual(pyvlx.config.port, 51200)

    def test_config_file_with_port(self) -> None:
        """Test host/password configuration via config.yaml."""
        pyvlx = PyVLX(loop=self.mocked_loop, path="test/data/test_config_with_port.yaml")  # pylint: disable=E1123
        self.assertEqual(pyvlx.config.port, 1234)

    def test_config_explicit(self) -> None:
        """Test host/password configuration via parameter."""
        pyvlx = PyVLX(loop=self.mocked_loop, host="192.168.2.127", password="velux123")  # pylint: disable=E1123
        self.assertEqual(pyvlx.config.password, "velux123")
        self.assertEqual(pyvlx.config.host, "192.168.2.127")
        self.assertEqual(pyvlx.config.port, 51200)

    def test_config_wrong1(self) -> None:
        """Test configuration with missing password."""
        with self.assertRaises(PyVLXException):
            PyVLX(loop=self.mocked_loop, path="test/data/test_config_wrong1.yaml")  # pylint: disable=E1123

    def test_config_wrong2(self) -> None:
        """Test configuration with missing host."""
        with self.assertRaises(PyVLXException):
            PyVLX(loop=self.mocked_loop, path="test/data/test_config_wrong2.yaml")  # pylint: disable=E1123

    def test_config_wrong3(self) -> None:
        """Test configuration with missing config node."""
        with self.assertRaises(PyVLXException):
            PyVLX(loop=self.mocked_loop, path="test/data/test_config_wrong3.yaml")  # pylint: disable=E1123

    def test_config_non_existent(self) -> None:
        """Test non existing configuration path."""
        with self.assertRaises(PyVLXException):
            PyVLX(loop=self.mocked_loop, path="test/data/test_config_non_existent.yaml")  # pylint: disable=E1123



"""Unit tests for FrameSetUTCRequest."""
import unittest
from datetime import datetime

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameSetUTCRequest


class TestFrameSetUTCRequest(unittest.TestCase):
    """Test class FrameSetUTCRequest."""

    # pylint: disable=too-many-public-methods,invalid-name

    EXAMPLE_FRAME = b"\x00\x07 \x00[\xfd\xaanE"
    EXAMPLE_TS = 1543350894

    def test_bytes(self):
        """Test FrameSetUTCRequest with NO_TYPE."""
        frame = FrameSetUTCRequest(timestamp=self.EXAMPLE_TS)
        self.assertEqual(bytes(frame), self.EXAMPLE_FRAME)

    def test_frame_from_raw(self):
        """Test parse FrameSetUTCRequest from raw."""
        frame = frame_from_raw(self.EXAMPLE_FRAME)
        self.assertTrue(isinstance(frame, FrameSetUTCRequest))
        self.assertEqual(frame.timestamp, self.EXAMPLE_TS)

    def test_str(self):
        """Test string representation of FrameSetUTCRequest."""
        frame = FrameSetUTCRequest(timestamp=1543350894)
        test_ts = datetime.fromtimestamp(1543350894).strftime("%Y-%m-%d %H:%M:%S")
        self.assertEqual(str(frame), '<FrameSetUTCRequest time="{}"/>'.format(test_ts))



"""Unit tests for FrameGetVersionRequest."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameGetVersionRequest


class TestFrameGetVersionRequest(unittest.TestCase):
    """Test class FrameGetVersionRequest."""

    # pylint: disable=too-many-public-methods,invalid-name

    EXAMPLE_FRAME = b"\x00\x03\x00\x08\x0b"

    def test_bytes(self):
        """Test FrameGetVersionRequest with NO_TYPE."""
        frame = FrameGetVersionRequest()
        self.assertEqual(bytes(frame), self.EXAMPLE_FRAME)

    def test_frame_from_raw(self):
        """Test parse FrameGetVersionRequest from raw."""
        frame = frame_from_raw(self.EXAMPLE_FRAME)
        self.assertTrue(isinstance(frame, FrameGetVersionRequest))

    def test_str(self):
        """Test string representation of FrameGetVersionRequest."""
        frame = FrameGetVersionRequest()
        self.assertEqual(str(frame), "<FrameGetVersionRequest/>")



"""Unit tests for FrameGetLocalTimeRequest."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameGetLocalTimeRequest


class TestFrameGetLocalTimeRequest(unittest.TestCase):
    """Test class for FrameGetLocalTimeRequest."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_bytes(self):
        """Test FrameGetLocalTimeRequest with NO_TYPE."""
        frame = FrameGetLocalTimeRequest()
        self.assertEqual(bytes(frame), b"\x00\x03\x20\x04\x27")

    def test_frame_from_raw(self):
        """Test parse FrameGetLocalTimeRequest from raw."""
        frame = frame_from_raw(b"\x00\x03\x20\x04\x27")
        self.assertTrue(isinstance(frame, FrameGetLocalTimeRequest))

    def test_str(self):
        """Test string representation of FrameGetLocalTimeRequest."""
        frame = FrameGetLocalTimeRequest()
        self.assertEqual(str(frame), "<FrameGetLocalTimeRequest/>")



"""Module for scene."""


class Scene:
    """Object for scene."""

    def __init__(self, pyvlx, ident, name):
        """Initialize Scene object."""
        self.pyvlx = pyvlx
        self.ident = ident
        self.name = name

    @classmethod
    def from_config(cls, pyvlx, item):
        """Read scene from configuration."""
        name = item['name']
        ident = item['id']
        return cls(pyvlx, ident, name)

    async def run(self):
        """Run scene."""
        await self.pyvlx.interface.api_call('scenes', 'run', {'id': self.ident})

    def get_name(self):
        """Return name of object."""
        return self.name

    def __str__(self) -> str:
        """Return object as readable string."""
        return '<Scene name="{0}" ' \
            'id="{1}" />' \
            .format(
                self.name,
                self.ident)

    def __eq__(self, other):
        """Equal operator."""
        return self.__dict__ == other.__dict__



"""Just a demo of the PyVLX module."""
import asyncio
import logging

from pyvlx import Position, PyVLX


async def main(loop):
    """Demonstrate functionality of PyVLX."""
    pyvlx = PyVLX('pyvlx.yaml', loop=loop)
    # Alternative:
    # pyvlx = PyVLX(host="192.168.2.127", password="velux123", loop=loop)

    # Runing scenes:
    await pyvlx.load_scenes()
    await pyvlx.scenes["All Windows Closed"].run()

    # Changing position of windows:
    await pyvlx.load_nodes()
    await pyvlx.nodes['Bath'].open()
    await pyvlx.nodes['Bath'].close()
    await pyvlx.nodes['Bath'].set_position(Position(position_percent=45))

    # Changing of on-off switches:
    # await pyvlx.nodes['CoffeeMaker'].set_on()
    # await pyvlx.nodes['CoffeeMaker'].set_off()

    # You can easily rename nodes:
    # await pyvlx.nodes["Window 10"].rename("Window 11")

    await pyvlx.disconnect()

if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)

    myloop = asyncio.new_event_loop()
    myloop.run_until_complete(main(myloop))
    # loop.run_forever()
    myloop.close()



"""Module for exceptions."""


class PyVLXException(Exception):
    """Exception class for PyVLX library."""

    def __init__(self, description):
        """Initialize exception with the given error message."""
        super(PyVLXException, self).__init__()
        self.description = description

    def __str__(self) -> str:
        """Return object as readable string."""
        return '<PyVLXException description="{0}" />' \
            .format(self.description)


class InvalidToken(PyVLXException):
    """KLF 200 token invalid or expired."""

    def __init__(self, error_code):
        """Initialize exception with the given error message."""
        super(InvalidToken, self).__init__("Invalid Token")
        self.error_code = error_code



"""Module for Opening devices."""
import asyncio
import datetime
from asyncio import Task
from typing import TYPE_CHECKING, Any, Optional

from .api.command_send import CommandSend
from .api.get_limitation import GetLimitation
from .const import Velocity
from .exception import PyVLXException
from .node import Node
from .parameter import (
    CurrentPosition, DualRollerShutterPosition, IgnorePosition, Parameter,
    Position, TargetPosition)

if TYPE_CHECKING:
    from pyvlx import PyVLX


class OpeningDevice(Node):
    """Meta class for opening device with one main parameter for position."""

    def __init__(
        self,
        pyvlx: "PyVLX",
        node_id: int,
        name: str,
        serial_number: Optional[str] = None,
        position_parameter: Parameter = Parameter(),
    ):
        """Initialize opening device.

        Parameters:
            * pyvlx: PyVLX object
            * node_id: internal id for addressing nodes.
                Provided by KLF 200 device
            * name: node name
            * serial_number: serial number of the node.
            * position_parameter: initial position of the opening device.

        """
        super().__init__(
            pyvlx=pyvlx, node_id=node_id, name=name, serial_number=serial_number
        )
        self.position: Position = Position(parameter=position_parameter)
        self.target: Position = Position(parameter=position_parameter)
        self.is_opening: bool = False
        self.is_closing: bool = False
        self.state_received_at: Optional[datetime.datetime] = None
        self.estimated_completion: Optional[datetime.datetime] = None
        self.use_default_velocity: bool = False
        self.default_velocity: Velocity = Velocity.DEFAULT
        self.open_position_target: int = 0
        self.close_position_target: int = 100
        self._update_task: Task | None = None

    async def _update_calls(self) -> None:
        """While cover are moving, perform periodically update calls."""
        while self.is_moving():
            await asyncio.sleep(1)
            await self.after_update()
        if self._update_task:
            self._update_task.cancel()
            self._update_task = None

    async def set_position(
        self,
        position: Position,
        velocity: Velocity | int | None = Velocity.DEFAULT,
        wait_for_completion: bool = True,
    ) -> None:
        """Set opening device to desired position.

        Parameters:
            * position: Position object containing the target position.
            * velocity: Velocity to be used during transition.
            * wait_for_completion: If set, function will return
                after device has reached target position.

        """
        kwargs: Any = {}

        if (
            velocity is None or velocity is Velocity.DEFAULT
        ) and self.use_default_velocity:
            velocity = self.default_velocity

        if isinstance(velocity, Velocity):
            if velocity is not Velocity.DEFAULT:
                if velocity is Velocity.SILENT:
                    kwargs["fp1"] = Parameter(raw=b"\x00\x00")
                else:
                    kwargs["fp1"] = Parameter(raw=b"\xC8\x00")
        elif isinstance(velocity, int):
            kwargs["fp1"] = Position.from_percent(velocity)

        command = CommandSend(
            pyvlx=self.pyvlx,
            wait_for_completion=wait_for_completion,
            node_id=self.node_id,
            parameter=position,
            **kwargs,
        )
        await command.send()
        await self.after_update()

    async def open(
        self,
        velocity: Velocity | int | None = Velocity.DEFAULT,
        wait_for_completion: bool = True,
    ) -> None:
        """Open opening device.

        Parameters:
            * velocity: Velocity to be used during transition.
            * wait_for_completion: If set, function will return
                after device has reached target position.

        """
        await self.set_position(
            position=Position(position_percent=self.open_position_target),
            velocity=velocity,
            wait_for_completion=wait_for_completion,
        )

    async def close(
        self,
        velocity: Velocity | int | None = Velocity.DEFAULT,
        wait_for_completion: bool = True,
    ) -> None:
        """Close opening device.

        Parameters:
            * velocity: Velocity to be used during transition.
            * wait_for_completion: If set, function will return
                after device has reached target position.

        """
        await self.set_position(
            position=Position(position_percent=self.close_position_target),
            velocity=velocity,
            wait_for_completion=wait_for_completion,
        )

    async def stop(self, wait_for_completion: bool = True) -> None:
        """Stop opening device.

        Parameters:
            * wait_for_completion: If set, function will return
                after device has reached target position.

        """
        await self.set_position(
            position=CurrentPosition(), wait_for_completion=wait_for_completion
        )

    def is_moving(self) -> bool:
        """Return moving state of the cover."""
        return self.is_opening or self.is_closing

    def movement_percent(self) -> int:
        """Return movement percentage of the cover."""
        if (
            self.estimated_completion is None
            or self.state_received_at is None
            or self.estimated_completion < datetime.datetime.now()
        ):
            return 100

        movement_duration_s: float = (
            self.estimated_completion - self.state_received_at
        ).total_seconds()
        time_passed_s: float = (
            datetime.datetime.now() - self.state_received_at
        ).total_seconds()

        percent: int = int(time_passed_s / movement_duration_s * 100)
        percent = max(percent, 0)
        percent = min(percent, 100)
        return percent

    def get_position(self) -> Position:
        """Return position of the cover."""
        if self.is_moving():
            percent = self.movement_percent()
            movement_origin = self.position.position_percent
            movement_target = self.target.position_percent
            current_position = (
                movement_origin + (movement_target - movement_origin) / 100 * percent
            )
            if not self._update_task:
                self._update_task = self.pyvlx.loop.create_task(self._update_calls())
            return Position(position_percent=int(current_position))
        return self.position

    def __str__(self) -> str:
        """Return object as readable string."""
        return '<{} name="{}" node_id="{}" serial_number="{}" position="{}"/>'.format(
            type(self).__name__,
            self.name,
            self.node_id,
            self.serial_number,
            self.position,
        )


class Window(OpeningDevice):
    """Window object."""

    def __init__(
        self,
        pyvlx: "PyVLX",
        node_id: int,
        name: str,
        serial_number: Optional[str],
        position_parameter: Parameter = Parameter(),
        rain_sensor: bool = False,
    ):
        """Initialize Window class.

        Parameters:
            * pyvlx: PyVLX object
            * node_id: internal id for addressing nodes.
                Provided by KLF 200 device
            * name: node name
            * serial_number: serial number of the node.
            * position_parameter: initial position of the opening device.
            * rain_sensor: set if device is equipped with a
                rain sensor.

        """
        super().__init__(
            pyvlx=pyvlx,
            node_id=node_id,
            name=name,
            serial_number=serial_number,
            position_parameter=position_parameter,
        )
        self.rain_sensor = rain_sensor

    def __str__(self) -> str:
        """Return object as readable string."""
        return '<{} name="{}" node_id="{}" rain_sensor={} serial_number="{}" position="{}"/>'.format(
            type(self).__name__,
            self.name,
            self.node_id,
            self.rain_sensor,
            self.serial_number,
            self.position,
        )

    async def get_limitation(self) -> GetLimitation:
        """Return limitation."""
        get_limitation = GetLimitation(pyvlx=self.pyvlx, node_id=self.node_id)
        await get_limitation.do_api_call()
        if not get_limitation.success:
            raise PyVLXException("Unable to send command")
        return get_limitation


class Blind(OpeningDevice):
    """Blind objects."""

    def __init__(
        self,
        pyvlx: "PyVLX",
        node_id: int,
        name: str,
        serial_number: Optional[str],
        position_parameter: Parameter = Parameter(),
    ):
        """Initialize Blind class.

        Parameters:
            * pyvlx: PyVLX object
            * node_id: internal id for addressing nodes.
                Provided by KLF 200 device
            * name: node name

        """
        super().__init__(
            pyvlx=pyvlx,
            node_id=node_id,
            name=name,
            serial_number=serial_number,
            position_parameter=position_parameter,
        )
        self.orientation: Position = Position(position_percent=0)
        self.target_orientation: Position = TargetPosition()
        self.target_position: Position = TargetPosition()
        self.open_orientation_target: int = 50
        self.close_orientation_target: int = 100

    async def set_position_and_orientation(
        self,
        position: Position,
        wait_for_completion: bool = True,
        velocity: Velocity | int | None = None,
        orientation: Optional[Position] = None,
    ) -> None:
        """Set window to desired position.

        Parameters:
            * position: Position object containing the current position.
            * velocity: Velocity to be used during transition.
            * target_position: Position object holding the target position
                which allows to adjust the position while the blind is in movement
                without stopping the blind (if orientation position has been changed.)
            * wait_for_completion: If set, function will return
                after device has reached target position.
            * orientation: If set, the orientation of the device will be set in the same request.
                Note, that, if the position is set to 0, the orientation will be set to 0 too.

        """
        self.target_position = position
        kwargs: Any = {}

        if orientation is not None:
            kwargs["fp3"] = orientation
        elif self.target_position == Position(position_percent=0):
            kwargs["fp3"] = Position(position_percent=0)
        else:
            kwargs["fp3"] = IgnorePosition()

        if (
            velocity is None or velocity is Velocity.DEFAULT
        ) and self.use_default_velocity:
            velocity = self.default_velocity

        if isinstance(velocity, Velocity):
            if velocity is not Velocity.DEFAULT:
                if velocity is Velocity.SILENT:
                    # The above code is declaring a variable called `kwargs`.
                    kwargs["fp1"] = Parameter(raw=b"\x00\x00")
                else:
                    kwargs["fp1"] = Parameter(raw=b"\xC8\x00")
        elif isinstance(velocity, int):
            kwargs["fp1"] = Position.from_percent(velocity)

        command = CommandSend(
            pyvlx=self.pyvlx,
            node_id=self.node_id,
            parameter=position,
            wait_for_completion=wait_for_completion,
            **kwargs
        )
        await command.send()
        await self.after_update()

    async def set_position(
        self,
        position: Position,
        velocity: Velocity | int | None = Velocity.DEFAULT,
        wait_for_completion: bool = True,
    ) -> None:
        """Set window to desired position.

        Parameters:
            * position: Position object containing the current position.
            * velocity: Velocity to be used during transition.
            * target_position: Position object holding the target position
                which allows to adjust the position while the blind is in movement
                without stopping the blind (if orientation position has been changed.)
            * wait_for_completion: If set, function will return
                after device has reached target position.
        """
        await self.set_position_and_orientation(
            position=position,
            wait_for_completion=wait_for_completion,
            velocity=velocity,
        )

    async def open(
        self,
        velocity: Velocity | int | None = Velocity.DEFAULT,
        wait_for_completion: bool = True,
    ) -> None:
        """Open window.

        Parameters:
            * velocity: Velocity to be used during transition.
            * wait_for_completion: If set, function will return
                after device has reached target position.
        """
        await self.set_position(
            position=Position(position_percent=self.open_position_target),
            velocity=velocity,
            wait_for_completion=wait_for_completion,
        )

    async def close(
        self,
        velocity: Velocity | int | None = Velocity.DEFAULT,
        wait_for_completion: bool = True,
    ) -> None:
        """Close window.

        Parameters:
            * velocity: Velocity to be used during transition.
            * wait_for_completion: If set, function will return
                after device has reached target position.
        """
        await self.set_position(
            position=Position(position_percent=self.close_position_target),
            velocity=velocity,
            wait_for_completion=wait_for_completion,
        )

    async def stop(self, wait_for_completion: bool = True) -> None:
        """Stop Blind position."""
        await self.set_position_and_orientation(
            position=CurrentPosition(),
            wait_for_completion=wait_for_completion,
            orientation=self.target_orientation,
        )

    async def set_orientation(
        self, orientation: Position, wait_for_completion: bool = True
    ) -> None:
        """Set Blind shades to desired orientation.

        Parameters:
            * orientation: Position object containing the target orientation.
            + target_orientation: Position object holding the target orientation
                which allows to adjust the orientation while the blind is in movement
                without stopping the blind (if the position has been changed.)
            * wait_for_completion: If set, function will return
                after device has reached target position.

        """
        self.target_orientation = orientation
        self.orientation = orientation

        fp3 = (
            Position(position_percent=0)
            if self.target_position == Position(position_percent=0)
            else self.target_orientation
        )

        print("Orientation in device: %s " % (orientation))
        command = CommandSend(
            pyvlx=self.pyvlx,
            wait_for_completion=wait_for_completion,
            node_id=self.node_id,
            parameter=self.target_position,
            fp3=fp3,
        )
        await command.send()
        await self.after_update()
        # KLF200 always send UNKNOWN position for functional parameter,
        # so orientation is set directly and not via GW_NODE_STATE_POSITION_CHANGED_NTF

    async def open_orientation(self, wait_for_completion: bool = True) -> None:
        """Open Blind slats orientation.

        Blind slats with ±90° orientation are open at 50%
        """
        await self.set_orientation(
            orientation=Position(position_percent=self.open_orientation_target),
            wait_for_completion=wait_for_completion,
        )

    async def close_orientation(self, wait_for_completion: bool = True) -> None:
        """Close Blind slats."""
        await self.set_orientation(
            orientation=Position(position_percent=self.close_orientation_target),
            wait_for_completion=wait_for_completion,
        )

    async def stop_orientation(self, wait_for_completion: bool = True) -> None:
        """Stop Blind slats."""
        await self.set_orientation(
            orientation=CurrentPosition(), wait_for_completion=wait_for_completion
        )


class Awning(OpeningDevice):
    """Awning objects."""


class DualRollerShutter(OpeningDevice):
    """DualRollerShutter object."""

    def __init__(
        self,
        pyvlx: "PyVLX",
        node_id: int,
        name: str,
        serial_number: Optional[str],
        position_parameter: Parameter = Parameter(),
    ):
        """Initialize Blind class.

        Parameters:
            * pyvlx: PyVLX object
            * node_id: internal id for addressing nodes.
                Provided by KLF 200 device
            * name: node name

        """
        super().__init__(
            pyvlx=pyvlx,
            node_id=node_id,
            name=name,
            serial_number=serial_number,
            position_parameter=position_parameter,
        )
        self.position_upper_curtain: Position = Position(position_percent=0)
        self.position_lower_curtain: Position = Position(position_percent=0)
        self.target_position: Any = Position()
        self.active_parameter: int = 0

    async def set_position(
        self,
        position: Position,
        velocity: Velocity | int | None = Velocity.DEFAULT,
        wait_for_completion: bool = True,
        curtain: str = "dual",
    ) -> None:
        """Set DualRollerShutter to desired position.

        Parameters:
            * position: Position object containing the current position.
            * target_position: Position object holding the target position
                which allows to adjust the position while the blind is in movement
            * wait_for_completion: If set, function will return
                after device has reached target position.
        """
        kwargs: Any = {}

        if curtain == "upper":
            self.target_position = DualRollerShutterPosition()
            self.active_parameter = 1
            kwargs["fp1"] = position
            kwargs["fp2"] = TargetPosition()
        elif curtain == "lower":
            self.target_position = DualRollerShutterPosition()
            self.active_parameter = 2
            kwargs["fp1"] = TargetPosition()
            kwargs["fp2"] = position
        else:
            self.target_position = position
            self.active_parameter = 0

        if (
            velocity is None or velocity is Velocity.DEFAULT
        ) and self.use_default_velocity:
            velocity = self.default_velocity

        if isinstance(velocity, Velocity):
            if velocity is not Velocity.DEFAULT:
                if velocity is Velocity.SILENT:
                    kwargs["fp3"] = Parameter(raw=b"\x00\x00")
                else:
                    kwargs["fp3"] = Parameter(raw=b"\xC8\x00")
        elif isinstance(velocity, int):
            kwargs["fp3"] = Position.from_percent(velocity)

        command = CommandSend(
            pyvlx=self.pyvlx,
            wait_for_completion=wait_for_completion,
            node_id=self.node_id,
            parameter=self.target_position,
            active_parameter=self.active_parameter,
            **kwargs
        )
        await command.send()
        if position.position <= Position.MAX:
            if curtain == "upper":
                self.position_upper_curtain = position
            elif curtain == "lower":
                self.position_lower_curtain = position
            else:
                self.position = position
        await self.after_update()

    async def open(
        self,
        velocity: Velocity | int | None = Velocity.DEFAULT,
        wait_for_completion: bool = True,
        curtain: str = "dual",
    ) -> None:
        """Open DualRollerShutter.

        Parameters:
            * wait_for_completion: If set, function will return
                after device has reached target position.

        """
        await self.set_position(
            position=Position(position_percent=self.open_position_target),
            velocity=velocity,
            wait_for_completion=wait_for_completion,
            curtain=curtain,
        )

    async def close(
        self,
        velocity: Velocity | int | None = Velocity.DEFAULT,
        wait_for_completion: bool = True,
        curtain: str = "dual",
    ) -> None:
        """Close DualRollerShutter.

        Parameters:
            * wait_for_completion: If set, function will return
                after device has reached target position.
        """
        await self.set_position(
            position=Position(position_percent=self.close_position_target),
            velocity=velocity,
            wait_for_completion=wait_for_completion,
            curtain=curtain,
        )

    async def stop(
        self,
        wait_for_completion: bool = True,
        velocity: Velocity | int | None = Velocity.DEFAULT,
        curtain: str = "dual",
    ) -> None:
        """Stop Blind position."""
        await self.set_position(
            position=CurrentPosition(),
            velocity=velocity,
            wait_for_completion=wait_for_completion,
            curtain=curtain,
        )


class RollerShutter(OpeningDevice):
    """RollerShutter object."""


class GarageDoor(OpeningDevice):
    """GarageDoor object."""


class Gate(OpeningDevice):
    """Gate object."""


class Blade(OpeningDevice):
    """Blade object."""



"""Unit tests for FrameGetLimitationStatusNotification."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames.frame_get_limitation import (
    FrameGetLimitationStatusNotification)
from pyvlx.const import Originator


class TestFrameGetLimitationStatusNotification(unittest.TestCase):
    """Test class for TestFrameGetLimitationStatusNotification."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_bytes(self):
        """Test FrameGetLimitationStatusNotification bytes."""
        frame = FrameGetLimitationStatusNotification()
        frame.session_id = 1
        frame.node_id = 1
        frame.parameter_id = 0
        frame.min_value = int.to_bytes(47668, 2, byteorder='big')
        frame.max_value = int.to_bytes(63487, 2, byteorder='big')
        frame.limit_originator = Originator.USER
        frame.limit_time = 0
        self.assertEqual(bytes(frame), b'\x00\r\x03\x14\x00\x01\x01\x00\xba4\xf7\xff\x01\x00\x9d')

    def test_frame_from_raw(self):
        """Test parse FrameGetLimitationStatusNotification from raw."""
        frame = frame_from_raw(b'\x00\r\x03\x14\x00\x01\x01\x00\xba4\xf7\xff\x01\x00\x9d')
        self.assertTrue(isinstance(frame, FrameGetLimitationStatusNotification))
        self.assertEqual(frame.limit_originator, Originator.USER)
        self.assertEqual(frame.node_id, 1)
        self.assertEqual(frame.parameter_id, 0)
        self.assertEqual(frame.session_id, 1)
        self.assertEqual(frame.max_value, b'\xf7')
        self.assertEqual(frame.min_value, b'\xba')
        self.assertEqual(frame.limit_time, 0)

    def test_str(self):
        """Test string representation of FrameGetLimitationStatusNotification."""
        frame = FrameGetLimitationStatusNotification()
        self.assertEqual(str(frame),
                         '<FrameGetLimitationStatusNotification node_id="0" '
                         'session_id="None" min_value="None" max_value="None" '
                         'originator="None" limit_time="None"/>')



"""Module for accessing KLF 200 gateway with python."""

from .dimmable_device import DimmableDevice, ExteriorHeating, Light, OnOffLight
from .discovery import VeluxDiscovery
from .exception import PyVLXException
from .klf200gateway import Klf200Gateway
from .log import PYVLXLOG
from .node import Node
from .nodes import Nodes
from .on_off_switch import OnOffSwitch
from .opening_device import (
    Awning, Blade, Blind, GarageDoor, Gate, OpeningDevice, RollerShutter,
    Window)
from .parameter import (
    CurrentIntensity, CurrentPosition, Intensity, Parameter, Position,
    SwitchParameter, SwitchParameterOff, SwitchParameterOn, UnknownIntensity,
    UnknownPosition)
# flake8: noqa
from .pyvlx import PyVLX
from .scene import Scene
from .scenes import Scenes



"""Unit test for Scenes object."""
import unittest
from unittest.mock import MagicMock

from pyvlx import PyVLX, Scene, Scenes

# pylint: disable=too-many-public-methods,invalid-name


class TestScenes(unittest.TestCase):
    """Test class for scenes object."""

    def test_get_item(self) -> None:
        """Test get_item."""
        pyvlx = MagicMock(spec=PyVLX)
        scenes = Scenes(pyvlx)
        scene1 = Scene(pyvlx, 0, "Scene_1")
        scenes.add(scene1)
        scene2 = Scene(pyvlx, 1, "Scene_2")
        scenes.add(scene2)
        scene3 = Scene(pyvlx, 4, "Scene_3")
        scenes.add(scene3)
        self.assertEqual(scenes["Scene_1"], scene1)
        self.assertEqual(scenes["Scene_2"], scene2)
        self.assertEqual(scenes["Scene_3"], scene3)
        self.assertEqual(scenes[0], scene1)
        self.assertEqual(scenes[1], scene2)
        self.assertEqual(scenes[4], scene3)

    def test_get_item_failed(self) -> None:
        """Test get_item with non existing object."""
        pyvlx = MagicMock(spec=PyVLX)
        scenes = Scenes(pyvlx)
        scene1 = Scene(pyvlx, 0, "Scene_1")
        scenes.add(scene1)
        with self.assertRaises(KeyError):
            scenes["Scene_2"]  # pylint: disable=pointless-statement
        with self.assertRaises(KeyError):
            scenes[1]  # pylint: disable=pointless-statement

    def test_iter(self) -> None:
        """Test iterator."""
        pyvlx = MagicMock(spec=PyVLX)
        scenes = Scenes(pyvlx)
        scene1 = Scene(pyvlx, 0, "Scene_1")
        scenes.add(scene1)
        scene2 = Scene(pyvlx, 1, "Scene_2")
        scenes.add(scene2)
        scene3 = Scene(pyvlx, 2, "Scene_3")
        scenes.add(scene3)
        self.assertEqual(tuple(scenes.__iter__()), (scene1, scene2, scene3))  # pylint: disable=unnecessary-dunder-call

    def test_len(self) -> None:
        """Test len."""
        pyvlx = MagicMock(spec=PyVLX)
        scenes = Scenes(pyvlx)
        self.assertEqual(len(scenes), 0)
        scenes.add(Scene(pyvlx, 0, "Scene_1"))
        scenes.add(Scene(pyvlx, 1, "Scene_2"))
        scenes.add(Scene(pyvlx, 2, "Scene_3"))
        scenes.add(Scene(pyvlx, 3, "Scene_4"))
        self.assertEqual(len(scenes), 4)

    def test_add_same_object(self) -> None:
        """Test adding object with same scene_id."""
        pyvlx = MagicMock(spec=PyVLX)
        scenes = Scenes(pyvlx)
        self.assertEqual(len(scenes), 0)
        scenes.add(Scene(pyvlx, 0, "Scene_1"))
        scenes.add(Scene(pyvlx, 1, "Scene_2"))
        scenes.add(Scene(pyvlx, 2, "Scene_3"))
        scenes.add(Scene(pyvlx, 1, "Scene_2_same_id"))
        self.assertEqual(len(scenes), 3)
        self.assertEqual(scenes[1].name, "Scene_2_same_id")

    def test_add_item_failed(self) -> None:
        """Test add() with wrong type."""
        pyvlx = MagicMock()
        scenes = Scenes(pyvlx)
        with self.assertRaises(TypeError):
            scenes.add(scenes)  # type: ignore
        with self.assertRaises(TypeError):
            scenes.add("scene")  # type: ignore

    def test_clear(self) -> None:
        """Test clear() method."""
        pyvlx = MagicMock(spec=PyVLX)
        scenes = Scenes(pyvlx)
        self.assertEqual(len(scenes), 0)
        scenes.add(Scene(pyvlx, 0, "Scene_1"))
        scenes.add(Scene(pyvlx, 1, "Scene_2"))
        scenes.clear()
        self.assertEqual(len(scenes), 0)



"""Module for retrieving limitation value from API."""

from typing import TYPE_CHECKING, Optional

from ..const import LimitationType, Originator
from ..parameter import Position
from .api_event import ApiEvent
from .frames import (
    FrameBase, FrameGetLimitationStatus, FrameGetLimitationStatusConfirmation,
    FrameGetLimitationStatusNotification)
from .session_id import get_new_session_id

if TYPE_CHECKING:
    from pyvlx import PyVLX


class GetLimitation(ApiEvent):
    """Class for retrieving gateway state from API."""

    def __init__(self, pyvlx: "PyVLX", node_id: int, limitation_type: LimitationType = LimitationType.MIN_LIMITATION):
        """Initialize SceneList class."""
        super().__init__(pyvlx=pyvlx)
        self.node_id = node_id
        self.limitation_type = limitation_type
        self.success = False
        self.notification_frame: Optional[FrameGetLimitationStatusNotification] = None
        self.session_id: Optional[int] = None
        self.min_value_raw: Optional[bytes] = None
        self.max_value_raw: Optional[bytes] = None
        self.originator: Optional[Originator] = None
        self.limit_time: Optional[int] = None

    @property
    def max_value(self) -> int:
        """Return max value."""
        assert self.max_value_raw is not None
        return Position.to_percent(self.max_value_raw)

    @property
    def min_value(self) -> int:
        """Return min value."""
        assert self.min_value_raw is not None
        return Position.to_percent(self.min_value_raw)

    async def handle_frame(self, frame: FrameBase) -> bool:
        """Handle incoming API frame, return True if this was the expected frame."""
        if isinstance(frame, FrameGetLimitationStatusConfirmation):
            return False  # Wait for Notification Frame
        if isinstance(frame, FrameGetLimitationStatusNotification):
            if frame.session_id == self.session_id:
                self.success = True
                self.min_value_raw = frame.min_value
                self.max_value_raw = frame.max_value
                self.originator = frame.limit_originator
                self.limit_time = frame.limit_time
                self.notification_frame = frame
                return True
        return False

    def request_frame(self) -> FrameGetLimitationStatus:
        """Construct initiating frame."""
        self.session_id = get_new_session_id()
        return FrameGetLimitationStatus(node_ids=[self.node_id], session_id=self.session_id,
                                        limitation_type=self.limitation_type)



"""Test for Position class."""
import unittest

from pyvlx import (
    CurrentPosition, Parameter, Position, SwitchParameter, SwitchParameterOff,
    SwitchParameterOn, UnknownPosition)
from pyvlx.exception import PyVLXException


class TestPosition(unittest.TestCase):
    """Test class for Position class."""

    # pylint: disable=invalid-name

    def test_no_position(self) -> None:
        """Test empty Position object."""
        self.assertEqual(Position().raw, b"\xF7\xFF")

    def test_raw(self) -> None:
        """Test raw assignment."""
        self.assertEqual(Position(Parameter(raw=b"\x00\x00")).raw, b"\x00\x00")
        self.assertEqual(Position(Parameter(raw=b"\x0A\x05")).raw, b"\x0A\x05")
        self.assertEqual(Position(Parameter(raw=b"\xC8\x00")).raw, b"\xC8\x00")

    def test_position(self) -> None:
        """Test initialization with position value."""
        self.assertEqual(Position(position=0).position, 0)
        self.assertEqual(Position(position=10).position, 10)
        self.assertEqual(Position(position=1234).position, 1234)
        self.assertEqual(Position(position=12345).position, 12345)
        self.assertEqual(Position(position=51200).position, 51200)

    def test_percent(self) -> None:
        """Test initialization with percent value."""
        self.assertEqual(Position(position_percent=0).position_percent, 0)
        self.assertEqual(Position(position_percent=5).position_percent, 5)
        self.assertEqual(Position(position_percent=50).position_percent, 50)
        self.assertEqual(Position(position_percent=95).position_percent, 95)
        self.assertEqual(Position(position_percent=100).position_percent, 100)

    def test_conversion(self) -> None:
        """Test conversion from one form to other."""
        self.assertEqual(Position(Parameter(raw=b"\x0A\x05")).position, 2565)
        self.assertEqual(Position(position_percent=50).position, 25600)
        self.assertEqual(Position(position=12345).position_percent, 24)
        self.assertEqual(Position(position=51200).position_percent, 100)
        # test tolerance for devices that report positions only close to 100%
        self.assertEqual(Position(position=51160).position_percent, 100)

    def test_fallback_to_unknown(self) -> None:
        """Test fallback to unknown."""
        self.assertEqual(Parameter(raw=b"\xC8\x01"), Parameter(raw=Parameter.from_int(Parameter.UNKNOWN_VALUE)))
        self.assertEqual(Parameter(raw=b"\xC9\x00"), Parameter(raw=Parameter.from_int(Parameter.UNKNOWN_VALUE)))
        self.assertEqual(Parameter(raw=b"\xD8\x00"), Parameter(raw=Parameter.from_int(Parameter.UNKNOWN_VALUE)))
        self.assertEqual(Parameter(raw=b"\xfe\x01"), Parameter(raw=Parameter.from_int(Parameter.UNKNOWN_VALUE)))

    def test_exception(self) -> None:
        """Test wrong initialization of Position."""
        with self.assertRaises(PyVLXException):
            Position(position=2.5)  # type: ignore
        with self.assertRaises(PyVLXException):
            Position(position=-1)
        with self.assertRaises(PyVLXException):
            Position(position=51201)
        with self.assertRaises(PyVLXException):
            Position(position_percent=2.5)  # type: ignore
        with self.assertRaises(PyVLXException):
            Position(position_percent=-1)
        with self.assertRaises(PyVLXException):
            Position(position_percent=101)
        with self.assertRaises(PyVLXException):
            Parameter(raw=b"\x00\x00\x00")
        with self.assertRaises(PyVLXException):
            Parameter(raw="\x00\x00")  # type: ignore

    def test_known(self) -> None:
        """Test 'known' property."""
        self.assertTrue(Position(Parameter(raw=b"\x12\x00")).known)
        self.assertTrue(Position(Parameter(raw=b"\xC8\x00")).known)
        self.assertFalse(Position(Parameter(raw=b"\xF7\xFF")).known)

        # Well, not really know. But at least not unknown:
        self.assertTrue(Position(Parameter(raw=b"\xD2\x00")).known)

    def test_open_closed(self) -> None:
        """Test open and closed property."""
        position_open = Position(position_percent=0)
        self.assertFalse(position_open.closed)
        self.assertTrue(position_open.open)
        position_closed = Position(position_percent=100)
        self.assertTrue(position_closed.closed)
        self.assertFalse(position_closed.open)
        position_closed = Position(position=51200)
        self.assertTrue(position_closed.closed)
        self.assertFalse(position_closed.open)
        position_closed = Position(position=51160)
        self.assertTrue(position_closed.closed)
        self.assertFalse(position_closed.open)
        position_half = Position(position_percent=50)
        self.assertFalse(position_half.closed)
        self.assertFalse(position_half.open)

    def test_str(self) -> None:
        """Test string representation."""
        self.assertEqual(str(Position(Parameter(raw=b"\xF7\xFF"))), "UNKNOWN")
        self.assertEqual(str(Position(position_percent=50)), "50 %")

    def test_unknown_position_class(self) -> None:
        """Test UnknownPosition class."""
        self.assertEqual(UnknownPosition().raw, b"\xF7\xFF")

    def test_current_position_class(self) -> None:
        """Test CurrentPosition class."""
        self.assertEqual(CurrentPosition().raw, b"\xD2\x00")

    def test_on_off(self) -> None:
        """Test SwitchParameter parameter."""
        parameter = SwitchParameter()
        self.assertFalse(parameter.is_on())
        self.assertFalse(parameter.is_off())
        parameter.set_on()
        self.assertTrue(parameter.is_on())
        self.assertFalse(parameter.is_off())
        parameter.set_off()
        self.assertFalse(parameter.is_on())
        self.assertTrue(parameter.is_off())

    def test_parsing_on_off(self) -> None:
        """Test parsing OnOFf from raw."""
        parameter_on = SwitchParameter(Parameter(raw=b"\x00\x00"))
        self.assertTrue(parameter_on.is_on())
        self.assertFalse(parameter_on.is_off())
        parameter_off = SwitchParameter(Parameter(raw=b"\xC8\x00"))
        self.assertFalse(parameter_off.is_on())
        self.assertTrue(parameter_off.is_off())

    def test_switch_parameter_on_class(self) -> None:
        """Test SwitchParameterOn class."""
        self.assertTrue(SwitchParameterOn().is_on())
        self.assertFalse(SwitchParameterOn().is_off())

    def test_switch_parameter_off_class(self) -> None:
        """Test SwitchParameterOff class."""
        self.assertFalse(SwitchParameterOff().is_on())
        self.assertTrue(SwitchParameterOff().is_off())



"""Unit tests for FrameGetProtocolVersionConfirmation."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameGetProtocolVersionConfirmation


class TestFrameGetProtocolVersionConfirmation(unittest.TestCase):
    """Test class for FrameGetProtocolVersionConfirmation."""

    # pylint: disable=too-many-public-methods,invalid-name

    EXAMPLE_FRAME = b"\x00\x07\x00\x0b\x04\xd2\x10\xe1+"

    def test_bytes(self):
        """Test FrameGetProtocolVersionConfirmation."""
        frame = FrameGetProtocolVersionConfirmation(
            major_version=1234, minor_version=4321
        )
        self.assertEqual(bytes(frame), self.EXAMPLE_FRAME)

    def test_frame_from_raw(self):
        """Test parse FrameGetProtocolVersionConfirmation from raw."""
        frame = frame_from_raw(self.EXAMPLE_FRAME)
        self.assertTrue(isinstance(frame, FrameGetProtocolVersionConfirmation))
        self.assertEqual(frame.major_version, 1234)
        self.assertEqual(frame.minor_version, 4321)

    def test_str(self):
        """Test string representation of FrameGetProtocolVersionConfirmation."""
        frame = FrameGetProtocolVersionConfirmation(
            major_version=1234, minor_version=4321
        )
        self.assertEqual(
            str(frame), '<FrameGetProtocolVersionConfirmation version="1234.4321"/>'
        )



"""Module for all KLF 200 API frames."""
# flake8: noqa

from .activate_scene import ActivateScene
from .api_event import ApiEvent
from .command_send import CommandSend
from .factory_default import FactoryDefault
from .get_all_nodes_information import GetAllNodesInformation
from .get_local_time import (
    FrameGetLocalTimeConfirmation, FrameGetLocalTimeRequest, GetLocalTime)
from .get_network_setup import GetNetworkSetup
from .get_node_information import GetNodeInformation
from .get_protocol_version import GetProtocolVersion
from .get_scene_list import GetSceneList
from .get_state import GetState
from .get_version import GetVersion
from .house_status_monitor import (
    HouseStatusMonitorDisable, HouseStatusMonitorEnable)
from .leave_learn_state import LeaveLearnState
from .password_enter import PasswordEnter
from .reboot import Reboot
from .set_node_name import SetNodeName
from .set_utc import SetUTC



"""Unit tests for FrameGetNodeInformationConfirmation."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameGetNodeInformationConfirmation


class TestFrameGetNodeInformationConfirmation(unittest.TestCase):
    """Test class for FrameGetNodeInformationConfirmation."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_bytes(self):
        """Test FrameGetNodeInformationConfirmation."""
        frame = FrameGetNodeInformationConfirmation(node_id=23)
        self.assertEqual(bytes(frame), b"\x00\x05\x02\x01\x00\x17\x11")

    def test_frame_from_raw(self):
        """Test parse FrameGetNodeInformationConfirmation from raw."""
        frame = frame_from_raw(b"\x00\x05\x02\x01\x00\x17\x11")
        self.assertTrue(isinstance(frame, FrameGetNodeInformationConfirmation))
        self.assertEqual(frame.node_id, 23)

    def test_str(self):
        """Test string representation of FrameGetNodeInformationConfirmation."""
        frame = FrameGetNodeInformationConfirmation(node_id=23)
        self.assertEqual(
            str(frame),
            '<FrameGetNodeInformationConfirmation node_id="23" status="NodeInformationStatus.OK"/>',
        )



"""Unit tests for FrameGetNodeInformationNotification."""
import unittest
from datetime import datetime

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameGetNodeInformationNotification
from pyvlx.api.frames.alias_array import AliasArray
from pyvlx.const import (
    NodeTypeWithSubtype, NodeVariation, OperatingState, Velocity)
from pyvlx.exception import PyVLXException
from pyvlx.parameter import Position


class TestFrameGetNodeInformationNotification(unittest.TestCase):
    """Test class for FrameGetNodeInformationNotification."""

    # pylint: disable=too-many-public-methods,invalid-name

    EXAMPLE_FRAME = (
        b"\x00\x7f\x02\x10\x17\x04\xd2\x02Fnord23\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x01\x00@\x17\r\x01\x01\x07\x01"
        b"\x02\x03\x04\x05\x06\x06\x08\x01\x00\x0c\x00{\x04\xd2\t)\r\x80"
        b"\x11\xd7\x00\x01\x03\x03\x02\x03\x0501234567890123456789\x65"
    )

    EXAMPLE_FRAME_UNKNOWN_TYPE = (
        b"\x00\x7f\x02\x10\x17\x04\xd2\x02Fnord23\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x01\x99\x99\x17\r\x01\x01\x07\x01"
        b"\x02\x03\x04\x05\x06\x06\x08\x01\x00\x0c\x00{\x04\xd2\t)\r\x80"
        b"\x11\xd7\x00\x01\x03\x03\x02\x03\x0501234567890123456789\x25"
    )

    def test_bytes(self) -> None:
        """Test FrameGetNodeInformationNotification."""
        frame = FrameGetNodeInformationNotification()
        frame.node_id = 23
        frame.order = 1234
        frame.placement = 2
        frame.name = "Fnord23"
        frame.velocity = Velocity.SILENT
        frame.node_type = NodeTypeWithSubtype.INTERIOR_VENETIAN_BLIND
        frame.product_group = 23
        frame.product_type = 13
        frame.node_variation = NodeVariation.TOPHUNG
        frame.power_mode = 1
        frame.build_number = 7
        frame.serial_number = "01:02:03:04:05:06:06:08"
        frame.state = OperatingState.ERROR_EXECUTING
        frame.current_position = Position(position=12)
        frame.target = Position(position=123)
        frame.current_position_fp1 = Position(position=1234)
        frame.current_position_fp2 = Position(position=2345)
        frame.current_position_fp3 = Position(position=3456)
        frame.current_position_fp4 = Position(position=4567)
        frame.remaining_time = 1
        frame.timestamp = 50528771
        frame.alias_array = AliasArray(raw=b"\x0501234567890123456789")
        self.assertEqual(bytes(frame), self.EXAMPLE_FRAME)

    def test_frame_from_raw(self) -> None:
        """Test parse FrameGetNodeInformationNotification from raw."""
        frame = frame_from_raw(self.EXAMPLE_FRAME)
        self.assertTrue(isinstance(frame, FrameGetNodeInformationNotification))
        self.assertEqual(frame.node_id, 23)
        self.assertEqual(frame.order, 1234)
        self.assertEqual(frame.placement, 2)
        self.assertEqual(frame.name, "Fnord23")
        self.assertEqual(frame.velocity, Velocity.SILENT)
        self.assertEqual(frame.node_type, NodeTypeWithSubtype.INTERIOR_VENETIAN_BLIND)
        self.assertEqual(frame.product_group, 23)
        self.assertEqual(frame.product_type, 13)
        self.assertEqual(frame.node_variation, NodeVariation.TOPHUNG)
        self.assertEqual(frame.power_mode, 1)
        self.assertEqual(frame.build_number, 7)
        self.assertEqual(frame.serial_number, "01:02:03:04:05:06:06:08")
        self.assertEqual(frame.state, OperatingState.ERROR_EXECUTING)
        self.assertEqual(Position(frame.current_position).position, 12)
        self.assertEqual(Position(frame.target).position, 123)
        self.assertEqual(Position(frame.current_position_fp1).position, 1234)
        self.assertEqual(Position(frame.current_position_fp2).position, 2345)
        self.assertEqual(Position(frame.current_position_fp3).position, 3456)
        self.assertEqual(Position(frame.current_position_fp4).position, 4567)
        self.assertEqual(frame.remaining_time, 1)
        self.assertEqual(frame.timestamp, 50528771)
        self.assertEqual(
            str(frame.alias_array),
            "3031=3233, 3435=3637, 3839=3031, 3233=3435, 3637=3839",
        )

    def test_frame_from_raw_unknown_type(self) -> None:
        """Test parse FrameGetNodeInformationNotification from raw."""
        frame = frame_from_raw(self.EXAMPLE_FRAME_UNKNOWN_TYPE)
        self.assertTrue(isinstance(frame, FrameGetNodeInformationNotification))
        self.assertEqual(frame.node_id, 23)
        self.assertEqual(frame.order, 1234)
        self.assertEqual(frame.placement, 2)
        self.assertEqual(frame.name, "Fnord23")
        self.assertEqual(frame.velocity, Velocity.SILENT)
        self.assertEqual(frame.node_type, NodeTypeWithSubtype.NO_TYPE)
        self.assertEqual(frame.product_group, 23)
        self.assertEqual(frame.product_type, 13)
        self.assertEqual(frame.node_variation, NodeVariation.TOPHUNG)
        self.assertEqual(frame.power_mode, 1)
        self.assertEqual(frame.build_number, 7)
        self.assertEqual(frame.serial_number, "01:02:03:04:05:06:06:08")
        self.assertEqual(frame.state, OperatingState.ERROR_EXECUTING)
        self.assertEqual(Position(frame.current_position).position, 12)
        self.assertEqual(Position(frame.target).position, 123)
        self.assertEqual(Position(frame.current_position_fp1).position, 1234)
        self.assertEqual(Position(frame.current_position_fp2).position, 2345)
        self.assertEqual(Position(frame.current_position_fp3).position, 3456)
        self.assertEqual(Position(frame.current_position_fp4).position, 4567)
        self.assertEqual(frame.remaining_time, 1)
        self.assertEqual(frame.timestamp, 50528771)
        self.assertEqual(
            str(frame.alias_array),
            "3031=3233, 3435=3637, 3839=3031, 3233=3435, 3637=3839",
        )

    def test_str(self) -> None:
        """Test string representation of FrameGetNodeInformationNotification."""
        frame = frame_from_raw(self.EXAMPLE_FRAME)
        test_ts = datetime.fromtimestamp(50528771).strftime("%Y-%m-%d %H:%M:%S")
        self.assertEqual(
            str(frame),
            '<FrameGetNodeInformationNotification node_id="23" order="1234" placement="2" '
            'name="Fnord23" velocity="Velocity.SILENT" node_type="NodeTypeWithSubtype.INTERIOR_VENETIAN_BLIND" '
            'product_group="23" product_type="13" node_variation="NodeVariation.TOPHUNG" '
            'power_mode="1" build_number="7" serial_number="01:02:03:04:05:06:06:08" state="ERROR_EXECUTING" '
            'current_position="0 %" target="0 %" current_position_fp1="2 %" '
            'current_position_fp2="5 %" current_position_fp3="7 %" current_position_fp4="9 %" '
            'remaining_time="1" time="{}" '
            'alias_array="3031=3233, 3435=3637, 3839=3031, 3233=3435, 3637=3839"/>'.format(
                test_ts
            ),
        )

    def test_serial_number(self) -> None:
        """Test serial number property."""
        frame = FrameGetNodeInformationNotification()
        frame.serial_number = "01:02:03:04:05:06:06:08"
        self.assertEqual(frame.serial_number, "01:02:03:04:05:06:06:08")

    def test_serial_number_none(self) -> None:
        """Test serial number property with no value set."""
        frame = FrameGetNodeInformationNotification()
        frame.serial_number = None
        self.assertEqual(frame.serial_number, None)

    def test_serial_number_not_set(self) -> None:
        """Test serial number property with not set."""
        frame = FrameGetNodeInformationNotification()
        self.assertEqual(frame.serial_number, None)

    def test_wrong_serial_number(self) -> None:
        """Test setting a wrong serial number."""
        frame = FrameGetNodeInformationNotification()
        with self.assertRaises(PyVLXException):
            frame.serial_number = "01:02:03:04:05:06:06"



"""Unit test for Nodes object."""
import unittest
from unittest.mock import MagicMock

from pyvlx import Blind, Nodes, PyVLX, RollerShutter, Window
from pyvlx.connection import Connection
from pyvlx.node import Node

# pylint: disable=too-many-public-methods,invalid-name


class TestNodes(unittest.TestCase):
    """Test class for nodes object."""

    def setUp(self) -> None:
        """Set up TestNodes."""
        self.pyvlx = MagicMock(spec=PyVLX)
        connection = MagicMock(spec=Connection)
        self.pyvlx.attach_mock(mock=connection, attribute="connection")

    def test_get_item(self) -> None:
        """Test get_item."""
        nodes = Nodes(self.pyvlx)
        window = Window(self.pyvlx, 0, "Window", "aa:bb:aa:bb:aa:bb:aa:00")
        nodes.add(window)
        blind = Blind(self.pyvlx, 1, "Blind", "aa:bb:aa:bb:aa:bb:aa:01")
        nodes.add(blind)
        roller_shutter = RollerShutter(
            self.pyvlx, 4, "Roller Shutter", "aa:bb:aa:bb:aa:bb:aa:04"
        )
        nodes.add(roller_shutter)
        self.assertEqual(nodes["Window"], window)
        self.assertEqual(nodes["Blind"], blind)
        self.assertEqual(nodes["Roller Shutter"], roller_shutter)
        self.assertEqual(nodes[0], window)
        self.assertEqual(nodes[1], blind)
        self.assertEqual(nodes[4], roller_shutter)

    def test_get_item_failed(self) -> None:
        """Test get_item with non existing object."""
        nodes = Nodes(self.pyvlx)
        window1 = Window(self.pyvlx, 0, "Window_1", "aa:bb:aa:bb:aa:bb:aa:00")
        nodes.add(window1)
        with self.assertRaises(KeyError):
            nodes["Window_2"]  # pylint: disable=pointless-statement
        with self.assertRaises(KeyError):
            nodes[1]  # pylint: disable=pointless-statement

    def test_contains_item(self) -> None:
        """Test contains operator."""
        nodes = Nodes(self.pyvlx)
        window1 = Window(self.pyvlx, 23, "Window_1", "aa:bb:aa:bb:aa:bb:aa:23")
        nodes.add(window1)
        window2 = Window(self.pyvlx, 42, "Window_2", "aa:bb:aa:bb:aa:bb:aa:42")  # not added
        self.assertTrue("Window_1" in nodes)
        self.assertTrue(23 in nodes)
        self.assertTrue(window1 in nodes)
        self.assertFalse("Window_2" in nodes)
        self.assertFalse(42 in nodes)
        self.assertFalse(window2 in nodes)

    def test_iter(self) -> None:
        """Test iterator."""
        nodes = Nodes(self.pyvlx)
        window1 = Window(self.pyvlx, 0, "Window_1", "aa:bb:aa:bb:aa:bb:aa:00")
        nodes.add(window1)
        window2 = Window(self.pyvlx, 1, "Window_2", "aa:bb:aa:bb:aa:bb:aa:01")
        nodes.add(window2)
        window3 = Window(self.pyvlx, 2, "Window_3", "aa:bb:aa:bb:aa:bb:aa:02")
        nodes.add(window3)
        self.assertEqual(tuple(nodes.__iter__()), (window1, window2, window3))  # pylint: disable=unnecessary-dunder-call

    def test_len(self) -> None:
        """Test len."""
        nodes = Nodes(self.pyvlx)
        self.assertEqual(len(nodes), 0)
        nodes.add(Window(self.pyvlx, 0, "Window_1", "aa:bb:aa:bb:aa:bb:aa:00"))
        nodes.add(Window(self.pyvlx, 1, "Window_2", "aa:bb:aa:bb:aa:bb:aa:01"))
        nodes.add(Window(self.pyvlx, 2, "Window_3", "aa:bb:aa:bb:aa:bb:aa:02"))
        nodes.add(Window(self.pyvlx, 3, "Window_4", "aa:bb:aa:bb:aa:bb:aa:03"))
        self.assertEqual(len(nodes), 4)

    def test_add_same_object(self) -> None:
        """Test adding object with same node_id."""
        nodes = Nodes(self.pyvlx)
        self.assertEqual(len(nodes), 0)
        nodes.add(Window(self.pyvlx, 0, "Window_1", "aa:bb:aa:bb:aa:bb:aa:00"))
        nodes.add(Window(self.pyvlx, 1, "Window_2", "aa:bb:aa:bb:aa:bb:aa:01"))
        nodes.add(Window(self.pyvlx, 2, "Window_3", "aa:bb:aa:bb:aa:bb:aa:02"))
        nodes.add(Window(self.pyvlx, 1, "Window_2_same_id", "aa:bb:aa:bb:aa:bb:aa:01"))
        self.assertEqual(len(nodes), 3)
        self.assertEqual(nodes[1].name, "Window_2_same_id")

    def test_add_item_failed(self) -> None:
        """Test add() with wrong type."""
        nodes = Nodes(self.pyvlx)
        with self.assertRaises(TypeError):
            nodes.add(nodes)  # type: ignore
        with self.assertRaises(TypeError):
            nodes.add("device")  # type: ignore

    def test_clear(self) -> None:
        """Test clear() method."""
        nodes = Nodes(self.pyvlx)
        self.assertEqual(len(nodes), 0)
        nodes.add(Window(self.pyvlx, 0, "Window_1", "aa:bb:aa:bb:aa:bb:aa:00"))
        nodes.add(Window(self.pyvlx, 1, "Window_2", "aa:bb:aa:bb:aa:bb:aa:01"))
        nodes.clear()
        self.assertEqual(len(nodes), 0)

    def test_node_str(self) -> None:
        """Test string representation of node."""
        node = Node(
            pyvlx=self.pyvlx,
            node_id=23,
            name="Test Abstract Node",
            serial_number="aa:bb:aa:bb:aa:bb:aa:23",
        )
        self.assertEqual(
            str(node),
            '<Node name="Test Abstract Node" node_id="23" serial_number="aa:bb:aa:bb:aa:bb:aa:23"/>',
        )



"""Unit tests for FrameErrorNotification."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import ErrorType, FrameErrorNotification


class TestErrorNotification(unittest.TestCase):
    """Test class for FrameErrorNotification."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_bytes(self):
        """Test FrameErrorNotification with Light."""
        frame = FrameErrorNotification(error_type=ErrorType.ErrorOnFrameStructure)
        self.assertEqual(bytes(frame), b"\x00\x04\x00\x00\x02\x06")

    def test_frame_from_raw(self):
        """Test parse FrameErrorNotification from raw."""
        frame = frame_from_raw(b"\x00\x04\x00\x00\x02\x06")
        self.assertTrue(isinstance(frame, FrameErrorNotification))
        self.assertEqual(frame.error_type, ErrorType.ErrorOnFrameStructure)

    def test_str(self):
        """Test string representation of FrameErrorNotification."""
        frame = FrameErrorNotification(error_type=ErrorType.ErrorOnFrameStructure)
        self.assertEqual(
            str(frame),
            '<FrameErrorNotification error_type="ErrorType.ErrorOnFrameStructure"/>',
        )



"""Unit tests for FrameNodeInformationChangedNotification."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameNodeInformationChangedNotification
from pyvlx.const import NodeVariation


class TestFrameNodeInformationChangedNotification(unittest.TestCase):
    """Test class for FrameNodeInformationChangedNotification."""

    # pylint: disable=too-many-public-methods,invalid-name

    EXAMPLE_FRAME = (
        b"\x00H\x02\x0c\x17Fnord23\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x04"
        b"\xd2\x02\x01\xd4"
    )

    def test_bytes(self):
        """Test FrameNodeInformationChangedNotification."""
        frame = FrameNodeInformationChangedNotification()
        frame.node_id = 23
        frame.order = 1234
        frame.placement = 2
        frame.name = "Fnord23"
        frame.node_variation = NodeVariation.TOPHUNG
        self.assertEqual(bytes(frame), self.EXAMPLE_FRAME)

    def test_frame_from_raw(self):
        """Test parse FrameNodeInformationChangedNotification from raw."""
        frame = frame_from_raw(self.EXAMPLE_FRAME)
        self.assertTrue(isinstance(frame, FrameNodeInformationChangedNotification))
        self.assertEqual(frame.node_id, 23)
        self.assertEqual(frame.order, 1234)
        self.assertEqual(frame.placement, 2)
        self.assertEqual(frame.node_variation, NodeVariation.TOPHUNG)
        self.assertEqual(frame.name, "Fnord23")

    def test_str(self):
        """Test string representation of FrameNodeInformationChangedNotification."""
        frame = frame_from_raw(self.EXAMPLE_FRAME)
        self.assertEqual(
            str(frame),
            '<FrameNodeInformationChangedNotification node_id="23" name="Fnord23" '
            'order="1234" placement="2" node_variation="NodeVariation.TOPHUNG"/>',
        )



"""Test for Intensity class."""
import unittest

from pyvlx import CurrentIntensity, Intensity, Parameter, UnknownIntensity
from pyvlx.exception import PyVLXException


class TestIntensity(unittest.TestCase):
    """Test class for Intensity class."""

    # pylint: disable=invalid-name

    def test_no_intensity(self) -> None:
        """Test empty Intensity object."""
        self.assertEqual(Intensity().raw, b"\xF7\xFF")

    def test_raw(self) -> None:
        """Test raw assignment."""
        self.assertEqual(Intensity(Parameter(raw=b"\x00\x00")).raw, b"\x00\x00")
        self.assertEqual(Intensity(Parameter(raw=b"\x0A\x05")).raw, b"\x0A\x05")
        self.assertEqual(Intensity(Parameter(raw=b"\xC8\x00")).raw, b"\xC8\x00")

    def test_intensity(self) -> None:
        """Test initialization with intensity value."""
        self.assertEqual(Intensity(intensity=0).intensity, 0)
        self.assertEqual(Intensity(intensity=10).intensity, 10)
        self.assertEqual(Intensity(intensity=1234).intensity, 1234)
        self.assertEqual(Intensity(intensity=12345).intensity, 12345)
        self.assertEqual(Intensity(intensity=51200).intensity, 51200)

    def test_percent(self) -> None:
        """Test initialization with percent value."""
        # Verify that percentages work correctly: 0% = off, 100% = fully on
        self.assertEqual(Intensity(intensity_percent=0).intensity_percent, 0)
        self.assertEqual(Intensity(intensity_percent=5).intensity_percent, 5)
        self.assertEqual(Intensity(intensity_percent=50).intensity_percent, 50)
        self.assertEqual(Intensity(intensity_percent=95).intensity_percent, 95)
        self.assertEqual(Intensity(intensity_percent=100).intensity_percent, 100)

    def test_conversion(self) -> None:
        """Test conversion from one form to other."""
        # 0% = off = raw 200 (0xC8) in first byte
        intensity_off = Intensity(intensity_percent=0)
        self.assertEqual(intensity_off.raw, b"\xC8\x00")
        self.assertEqual(intensity_off.intensity, 51200)

        # 100% = fully on = raw 0 in first byte
        intensity_on = Intensity(intensity_percent=100)
        self.assertEqual(intensity_on.raw, b"\x00\x00")
        self.assertEqual(intensity_on.intensity, 0)

        # 50% = half brightness = raw 100 in first byte
        intensity_half = Intensity(intensity_percent=50)
        self.assertEqual(intensity_half.raw, b"\x64\x00")
        self.assertEqual(intensity_half.intensity, 25600)
        self.assertEqual(intensity_half.intensity_percent, 50)

        # Test reverse conversion from raw
        self.assertEqual(Intensity(Parameter(raw=b"\xC8\x00")).intensity_percent, 0)
        self.assertEqual(Intensity(Parameter(raw=b"\x00\x00")).intensity_percent, 100)
        self.assertEqual(Intensity(Parameter(raw=b"\x64\x00")).intensity_percent, 50)

    def test_exception(self) -> None:
        """Test wrong initialization of Intensity."""
        with self.assertRaises(PyVLXException):
            Intensity(intensity=2.5)  # type: ignore
        with self.assertRaises(PyVLXException):
            Intensity(intensity=-1)
        with self.assertRaises(PyVLXException):
            Intensity(intensity=51201)
        with self.assertRaises(PyVLXException):
            Intensity(intensity_percent=2.5)  # type: ignore
        with self.assertRaises(PyVLXException):
            Intensity(intensity_percent=-1)
        with self.assertRaises(PyVLXException):
            Intensity(intensity_percent=101)

    def test_known(self) -> None:
        """Test 'known' property."""
        self.assertTrue(Intensity(Parameter(raw=b"\x12\x00")).known)
        self.assertTrue(Intensity(Parameter(raw=b"\xC8\x00")).known)
        self.assertFalse(Intensity(Parameter(raw=b"\xF7\xFF")).known)

        # Well, not really known. But at least not unknown:
        self.assertTrue(Intensity(Parameter(raw=b"\xD2\x00")).known)

    def test_on_off(self) -> None:
        """Test on and off property."""
        # 100% intensity = fully on = FULL_BRIGHTNESS = 0x0000
        intensity_on = Intensity(intensity_percent=100)
        self.assertFalse(intensity_on.off)
        self.assertTrue(intensity_on.on)

        # 0% intensity = off = NO_BRIGHTNESS = 0xC800
        intensity_off = Intensity(intensity_percent=0)
        self.assertTrue(intensity_off.off)
        self.assertFalse(intensity_off.on)

        # From raw values
        intensity_on_raw = Intensity(intensity=0)
        self.assertTrue(intensity_on_raw.on)
        self.assertFalse(intensity_on_raw.off)

        intensity_off_raw = Intensity(intensity=51200)
        self.assertFalse(intensity_off_raw.on)
        self.assertTrue(intensity_off_raw.off)

        # Half intensity is neither fully on nor off
        intensity_half = Intensity(intensity_percent=50)
        self.assertFalse(intensity_half.off)
        self.assertFalse(intensity_half.on)

    def test_str(self) -> None:
        """Test string representation."""
        self.assertEqual(str(Intensity(Parameter(raw=b"\xF7\xFF"))), "UNKNOWN")
        self.assertEqual(str(Intensity(intensity_percent=50)), "50 %")
        self.assertEqual(str(Intensity(intensity_percent=0)), "0 %")
        self.assertEqual(str(Intensity(intensity_percent=100)), "100 %")

    def test_unknown_intensity_class(self) -> None:
        """Test UnknownIntensity class."""
        self.assertEqual(UnknownIntensity().raw, b"\xF7\xFF")

    def test_current_intensity_class(self) -> None:
        """Test CurrentIntensity class."""
        self.assertEqual(CurrentIntensity().raw, b"\xD2\x00")

    def test_percent_to_raw_mapping(self) -> None:
        """Test that percent correctly maps to raw values (inverted for Intensity)."""
        # For Intensity: 0% (off) should give raw value 200 (0xC8)
        self.assertEqual(Intensity.from_percent(0), b"\xC8\x00")
        # For Intensity: 100% (fully on) should give raw value 0
        self.assertEqual(Intensity.from_percent(100), b"\x00\x00")
        # For Intensity: 50% should give raw value 100 (0x64)
        self.assertEqual(Intensity.from_percent(50), b"\x64\x00")
        # For Intensity: 25% should give raw value 150 (0x96)
        self.assertEqual(Intensity.from_percent(25), b"\x96\x00")
        # For Intensity: 75% should give raw value 50 (0x32)
        self.assertEqual(Intensity.from_percent(75), b"\x32\x00")

    def test_raw_to_percent_mapping(self) -> None:
        """Test that raw values correctly map to percent (inverted for Intensity)."""
        # For Intensity: raw value 200 (0xC8) should give 0% (off)
        self.assertEqual(Intensity.to_percent(b"\xC8\x00"), 0)
        # For Intensity: raw value 0 should give 100% (fully on)
        self.assertEqual(Intensity.to_percent(b"\x00\x00"), 100)
        # For Intensity: raw value 100 (0x64) should give 50%
        self.assertEqual(Intensity.to_percent(b"\x64\x00"), 50)
        # For Intensity: raw value 150 (0x96) should give 25%
        self.assertEqual(Intensity.to_percent(b"\x96\x00"), 25)
        # For Intensity: raw value 50 (0x32) should give 75%
        self.assertEqual(Intensity.to_percent(b"\x32\x00"), 75)
